package generated.eTicket.javacard.eTicketCard;

import javacard.framework.ISO7816;
import javacard.framework.ISOException;
import javacard.framework.Util;
/**
 * Class for T(L)V encoding of messages and other classes
 * 
 */
public class Coding {

	/**
	 * offset into encodeDestination for writing/reading
	 */
	private short encoding_length;

	/**
	 * array used for encoding/decoding
	 */
	private byte[] encodeDestination;

	/**
	 * Singleton instance
	 */
	private static Coding the_instance;

	/**
	 * Singleton accessor
	 * 
	 * @return Singleton instance
	 */

	public static Coding getInstance() {
		if (the_instance == null)
			the_instance = new Coding();
		return the_instance;
	}
	/**
	 * return offset into encodeDestination
	 * 
	 * @return value of encoding_length
	 */
	public short getEncodingLength() {
		return encoding_length;
	}

	/**
	 * initializes encoding. This method must be called before encoding an object.
	 * The offset (encoding_length) is reset to 0 and encodeDestination is set to dest.
	 * 
	 * @param dest The (empty) destination array for encoding
	 */
	public void encodeInit(byte[] dest) {
		encodeDestination = dest;
		encoding_length = 0;
	}
	/**
	 * initializes decoding. The offset (encoding_length) is reset to 0.
	 */
	public void decodeInit() {
		encoding_length = 0;
	}

	/**
	 * initializes encoding and encodes an object.
	 * 
	 * @param o The object to encode
	 * @param dest destination array for the encoding
	 * @return length of the encoded object
	 */

	private byte decodeByte(byte[] in) {
		if (in[encoding_length++] != Code.BYTE) {

			SimpleComm.raiseError();
		}

		byte res = in[encoding_length++];
		return res;
	}

	private boolean decodeBoolean(byte[] in) {
		if (in[encoding_length++] != Code.BOOLEAN) {

			SimpleComm.raiseError();

		}
		boolean res = (in[encoding_length++] == 1 ? true : false);
		return res;
	}

	public short encode(Codeable o, byte[] dest) {
		encodeInit(dest);

		try {
			encode(o);
		}

		catch (ArrayIndexOutOfBoundsException e) {
			SimpleComm.raiseError();
		}
		return encoding_length;

	}

	/**
	 * computes the encoding length of an object without actually encoding it.
	 * 
	 * @param o The object of which the encoding length should be computed
	 * @return encoding length
	 */

	public short computeSize(Codeable o) {
		encoding_length = 0;
		try {
			compSize(o);
			return encoding_length;
		} catch (Exception e) {
			return -1;
		}
	}

	private void encodeShort(short o) {
		encodeDestination[encoding_length++] = Code.SHORT;
		Util.setShort(encodeDestination, encoding_length, o);
		encoding_length += 2;
	}

	public void encodeByte(byte o) {
		encodeDestination[encoding_length++] = Code.BYTE;
		encodeDestination[encoding_length++] = o;
	}

	private void encodeBoolean(boolean o) {
		encodeDestination[encoding_length++] = Code.BOOLEAN;
		encodeDestination[encoding_length++] = (byte) (o ? 1 : 0);
	}

	private void compSizeShort(short o) {
		encoding_length += 3;
	}

	private void compSizeByte(byte o) {
		encoding_length += 2;
	}

	private void compSizeBoolean(boolean o) {
		encoding_length += 2;
	}

	private short decodeShort(byte[] in) {
		if (in[encoding_length++] != Code.SHORT)
			SimpleComm.raiseError();
		short res = Util.getShort(in, encoding_length);
		encoding_length += 2;
		return res;
	}

	public void encodeByteArray(byte[] o) {

		encodeDestination[encoding_length++] = Code.BYTEARRAY;

		Util.setShort(encodeDestination, encoding_length, (short) (o.length));
		encoding_length += 2;

		Util.arrayCopy(o, (short) 0, encodeDestination, encoding_length,
				(short) o.length);

		encoding_length += o.length;
	}

	private void compSizeByteArray(byte[] o) {
		encoding_length += 3;
		encoding_length += o.length;
	}

	private void decodeByteArrayInto(byte[] in, byte[] into) {
		if (in[encoding_length++] != Code.BYTEARRAY)
			SimpleComm.raiseError();
		short len = Util.getShort(in, encoding_length);
		encoding_length += 2;
		Util.arrayCopy(in, encoding_length, into, (short) 0, len);
		encoding_length += len;
	}

	/**
	 * encodes an object without initializing destination and offset.
	 * 
	 * @param c object to encode
	 */

	public void encode(Codeable c) {

		switch (c.getCode()) {
			case Code.BUYTICKET :
				encodeBuyTicket((BuyTicket) c);
				break;
			case Code.REQTICKET :
				encodeReqTicket((ReqTicket) c);
				break;
			case Code.RESREQTICKET :
				encodeResReqTicket((ResReqTicket) c);
				break;
			case Code.DELETETICKET :
				encodeDeleteTicket((DeleteTicket) c);
				break;
			case Code.CLIENTHELLO :
				encodeClientHello((ClientHello) c);
				break;
			case Code.SESSIONKEY :
				encodeSessionKey((SessionKey) c);
				break;
			case Code.SERVERHELLO :
				encodeServerHello((ServerHello) c);
				break;
			case Code.HANDSHAKE :
				encodeHandshake((Handshake) c);
				break;
			case Code.RECOVERTICKET :
				encodeRecoverTicket((RecoverTicket) c);
				break;
			case Code.REQRECOVERY :
				encodeReqRecovery((ReqRecovery) c);
				break;
			case Code.RESREQRECOVERY :
				encodeResReqRecovery((ResReqRecovery) c);
				break;
			case Code.SHOWALLTICKETS :
				encodeShowAllTickets((ShowAllTickets) c);
				break;
			case Code.RETURNALLTICKETS :
				encodeReturnAllTickets((ReturnAllTickets) c);
				break;
			case Code.TICKETRECEIVED :
				encodeTicketReceived((TicketReceived) c);
				break;
			case Code.TICKETRECEIVEDRECOVERY :
				encodeTicketReceivedRecovery((TicketReceivedRecovery) c);
				break;
			case Code.CARDAUTH :
				encodeCardAuth((CardAuth) c);
				break;
			case Code.CARDAUTHRES :
				encodeCardAuthRes((CardAuthRes) c);
				break;
			case Code.STAMPTICKET :
				encodeStampTicket((StampTicket) c);
				break;
			case Code.STAMPTICKETSUCCESSFUL :
				encodeStampTicketSuccessful((StampTicketSuccessful) c);
				break;
			case Code.RETURNALLITICKETS :
				encodeReturnAllITickets((ReturnAllITickets) c);
				break;
			case Code.SHOWALLITICKETS :
				encodeShowAllITickets((ShowAllITickets) c);
				break;
			case Code.TICKETINFO :
				encodeTicketInfo((TicketInfo) c);
				break;
			case Code.SECRET :
				encodeSecret((Secret) c);
				break;
			case Code.ENCDATASYMM :
				encodeEncDataSymm((EncDataSymm) c);
				break;
			case Code.BUYINFO :
				encodeBuyInfo((BuyInfo) c);
				break;
			case Code.RESREQTICKETDATA :
				encodeResReqTicketData((ResReqTicketData) c);
				break;
			case Code.ID :
				encodeID((ID) c);
				break;
			case Code.PUBLICKEY :
				encodePublicKey((PublicKey) c);
				break;
			case Code.SIGNEDDATA :
				encodeSignedData((SignedData) c);
				break;
			case Code.ENCDATAASYMM :
				encodeEncDataAsymm((EncDataAsymm) c);
				break;
			case Code.CERTIFICATE :
				encodeCertificate((Certificate) c);
				break;
			case Code.CLIENTHELLOMSG :
				encodeClientHelloMsg((ClientHelloMsg) c);
				break;
			case Code.SESSIONKEYMSG :
				encodeSessionKeyMsg((SessionKeyMsg) c);
				break;
			case Code.SERVERHELLOMSG :
				encodeServerHelloMsg((ServerHelloMsg) c);
				break;
			case Code.RECOVERYINFO :
				encodeRecoveryInfo((RecoveryInfo) c);
				break;
			case Code.RECOVERYDATA :
				encodeRecoveryData((RecoveryData) c);
				break;
			case Code.LISTOFTICKET :
				encodeListOfTicket((ListOfTicket) c);
				break;
			case Code.TICKETRECEIVEDDATA :
				encodeTicketReceivedData((TicketReceivedData) c);
				break;
			case Code.TICKETRECEIVEDDATARECOVERY :
				encodeTicketReceivedDataRecovery((TicketReceivedDataRecovery) c);
				break;
			case Code.CARDAUTHRESMSG :
				encodeCardAuthResMsg((CardAuthResMsg) c);
				break;
			case Code.STAMPTICKETMSG :
				encodeStampTicketMsg((StampTicketMsg) c);
				break;
			case Code.RETURNALLITICKETSMSG :
				encodeReturnAllITicketsMsg((ReturnAllITicketsMsg) c);
				break;
			case Code.CCINFO :
				encodeCCInfo((CCInfo) c);
				break;
			case Code.NONCE :
				encodeNonce((Nonce) c);
				break;
			case Code.TICKET :
				encodeTicket((Ticket) c);
				break;
			case Code.SYMMKEY :
				encodeSymmKey((SymmKey) c);
				break;
			case Code.STAMP :
				encodeStamp((Stamp) c);
				break;
			case Code.DATETIME :
				encodeDateTime((DateTime) c);
				break;
			case Code.PRIVATEKEY :
				encodePrivateKey((PrivateKey) c);
				break;
			// never reached
			default :

				SimpleComm.raiseError();

		}
	}

	public Message decodeMessage(byte[] in, short offset, short expectedLength) {
		encoding_length = offset;
		Message m = null;
		try {
			switch (in[encoding_length]) {
				case Code.BUYTICKET :
					m = Store.newBuyTicket();
					decodeBuyTicketInto(in, (BuyTicket) m);
					break;
				case Code.REQTICKET :
					m = Store.newReqTicket();
					decodeReqTicketInto(in, (ReqTicket) m);
					break;
				case Code.RESREQTICKET :
					m = Store.newResReqTicket();
					decodeResReqTicketInto(in, (ResReqTicket) m);
					break;
				case Code.DELETETICKET :
					m = Store.newDeleteTicket();
					decodeDeleteTicketInto(in, (DeleteTicket) m);
					break;
				case Code.CLIENTHELLO :
					m = Store.newClientHello();
					decodeClientHelloInto(in, (ClientHello) m);
					break;
				case Code.SESSIONKEY :
					m = Store.newSessionKey();
					decodeSessionKeyInto(in, (SessionKey) m);
					break;
				case Code.SERVERHELLO :
					m = Store.newServerHello();
					decodeServerHelloInto(in, (ServerHello) m);
					break;
				case Code.HANDSHAKE :
					m = Store.newHandshake();
					decodeHandshakeInto(in, (Handshake) m);
					break;
				case Code.RECOVERTICKET :
					m = Store.newRecoverTicket();
					decodeRecoverTicketInto(in, (RecoverTicket) m);
					break;
				case Code.REQRECOVERY :
					m = Store.newReqRecovery();
					decodeReqRecoveryInto(in, (ReqRecovery) m);
					break;
				case Code.RESREQRECOVERY :
					m = Store.newResReqRecovery();
					decodeResReqRecoveryInto(in, (ResReqRecovery) m);
					break;
				case Code.SHOWALLTICKETS :
					m = Store.newShowAllTickets();
					decodeShowAllTicketsInto(in, (ShowAllTickets) m);
					break;
				case Code.RETURNALLTICKETS :
					m = Store.newReturnAllTickets();
					decodeReturnAllTicketsInto(in, (ReturnAllTickets) m);
					break;
				case Code.TICKETRECEIVED :
					m = Store.newTicketReceived();
					decodeTicketReceivedInto(in, (TicketReceived) m);
					break;
				case Code.TICKETRECEIVEDRECOVERY :
					m = Store.newTicketReceivedRecovery();
					decodeTicketReceivedRecoveryInto(in,
							(TicketReceivedRecovery) m);
					break;
				case Code.CARDAUTH :
					m = Store.newCardAuth();
					decodeCardAuthInto(in, (CardAuth) m);
					break;
				case Code.CARDAUTHRES :
					m = Store.newCardAuthRes();
					decodeCardAuthResInto(in, (CardAuthRes) m);
					break;
				case Code.STAMPTICKET :
					m = Store.newStampTicket();
					decodeStampTicketInto(in, (StampTicket) m);
					break;
				case Code.STAMPTICKETSUCCESSFUL :
					m = Store.newStampTicketSuccessful();
					decodeStampTicketSuccessfulInto(in,
							(StampTicketSuccessful) m);
					break;
				case Code.RETURNALLITICKETS :
					m = Store.newReturnAllITickets();
					decodeReturnAllITicketsInto(in, (ReturnAllITickets) m);
					break;
				case Code.SHOWALLITICKETS :
					m = Store.newShowAllITickets();
					decodeShowAllITicketsInto(in, (ShowAllITickets) m);
					break;
			}
		} catch (ArrayIndexOutOfBoundsException e) {
			SimpleComm.raiseError();
		}
		if (m == null || encoding_length != (short) (expectedLength + offset))
			SimpleComm.raiseError();
		return m;
	}

	public void compSize(Codeable c) {
		switch (c.getCode()) {
			case Code.BUYTICKET :

				compSizeBuyTicket((BuyTicket) c);
				break;
			case Code.REQTICKET :

				compSizeReqTicket((ReqTicket) c);
				break;
			case Code.RESREQTICKET :

				compSizeResReqTicket((ResReqTicket) c);
				break;
			case Code.DELETETICKET :

				compSizeDeleteTicket((DeleteTicket) c);
				break;
			case Code.CLIENTHELLO :

				compSizeClientHello((ClientHello) c);
				break;
			case Code.SESSIONKEY :

				compSizeSessionKey((SessionKey) c);
				break;
			case Code.SERVERHELLO :

				compSizeServerHello((ServerHello) c);
				break;
			case Code.HANDSHAKE :

				compSizeHandshake((Handshake) c);
				break;
			case Code.RECOVERTICKET :

				compSizeRecoverTicket((RecoverTicket) c);
				break;
			case Code.REQRECOVERY :

				compSizeReqRecovery((ReqRecovery) c);
				break;
			case Code.RESREQRECOVERY :

				compSizeResReqRecovery((ResReqRecovery) c);
				break;
			case Code.SHOWALLTICKETS :

				compSizeShowAllTickets((ShowAllTickets) c);
				break;
			case Code.RETURNALLTICKETS :

				compSizeReturnAllTickets((ReturnAllTickets) c);
				break;
			case Code.TICKETRECEIVED :

				compSizeTicketReceived((TicketReceived) c);
				break;
			case Code.TICKETRECEIVEDRECOVERY :

				compSizeTicketReceivedRecovery((TicketReceivedRecovery) c);
				break;
			case Code.CARDAUTH :

				compSizeCardAuth((CardAuth) c);
				break;
			case Code.CARDAUTHRES :

				compSizeCardAuthRes((CardAuthRes) c);
				break;
			case Code.STAMPTICKET :

				compSizeStampTicket((StampTicket) c);
				break;
			case Code.STAMPTICKETSUCCESSFUL :

				compSizeStampTicketSuccessful((StampTicketSuccessful) c);
				break;
			case Code.RETURNALLITICKETS :

				compSizeReturnAllITickets((ReturnAllITickets) c);
				break;
			case Code.SHOWALLITICKETS :

				compSizeShowAllITickets((ShowAllITickets) c);
				break;
			case Code.TICKETINFO :

				compSizeTicketInfo((TicketInfo) c);
				break;
			case Code.SECRET :

				compSizeSecret((Secret) c);
				break;
			case Code.ENCDATASYMM :

				compSizeEncDataSymm((EncDataSymm) c);
				break;
			case Code.BUYINFO :

				compSizeBuyInfo((BuyInfo) c);
				break;
			case Code.RESREQTICKETDATA :

				compSizeResReqTicketData((ResReqTicketData) c);
				break;
			case Code.ID :

				compSizeID((ID) c);
				break;
			case Code.PUBLICKEY :

				compSizePublicKey((PublicKey) c);
				break;
			case Code.SIGNEDDATA :

				compSizeSignedData((SignedData) c);
				break;
			case Code.ENCDATAASYMM :

				compSizeEncDataAsymm((EncDataAsymm) c);
				break;
			case Code.CERTIFICATE :

				compSizeCertificate((Certificate) c);
				break;
			case Code.CLIENTHELLOMSG :

				compSizeClientHelloMsg((ClientHelloMsg) c);
				break;
			case Code.SESSIONKEYMSG :

				compSizeSessionKeyMsg((SessionKeyMsg) c);
				break;
			case Code.SERVERHELLOMSG :

				compSizeServerHelloMsg((ServerHelloMsg) c);
				break;
			case Code.RECOVERYINFO :

				compSizeRecoveryInfo((RecoveryInfo) c);
				break;
			case Code.RECOVERYDATA :

				compSizeRecoveryData((RecoveryData) c);
				break;
			case Code.LISTOFTICKET :

				compSizeListOfTicket((ListOfTicket) c);
				break;
			case Code.TICKETRECEIVEDDATA :

				compSizeTicketReceivedData((TicketReceivedData) c);
				break;
			case Code.TICKETRECEIVEDDATARECOVERY :

				compSizeTicketReceivedDataRecovery((TicketReceivedDataRecovery) c);
				break;
			case Code.CARDAUTHRESMSG :

				compSizeCardAuthResMsg((CardAuthResMsg) c);
				break;
			case Code.STAMPTICKETMSG :

				compSizeStampTicketMsg((StampTicketMsg) c);
				break;
			case Code.RETURNALLITICKETSMSG :

				compSizeReturnAllITicketsMsg((ReturnAllITicketsMsg) c);
				break;
			case Code.CCINFO :

				compSizeCCInfo((CCInfo) c);
				break;
			case Code.NONCE :

				compSizeNonce((Nonce) c);
				break;
			case Code.TICKET :

				compSizeTicket((Ticket) c);
				break;
			case Code.SYMMKEY :

				compSizeSymmKey((SymmKey) c);
				break;
			case Code.STAMP :

				compSizeStamp((Stamp) c);
				break;
			case Code.DATETIME :

				compSizeDateTime((DateTime) c);
				break;
			case Code.PRIVATEKEY :

				compSizePrivateKey((PrivateKey) c);
				break;
			// never reached
			default :

				SimpleComm.raiseError();
				break;

		}
	}

	public PlainData decodePlainData(byte[] in, byte expectedType) {
		encoding_length = 0;
		PlainData p = null;
		try {
			if (in[encoding_length] != expectedType) {

				SimpleComm.raiseError();
			}
			switch (in[encoding_length]) {

				case Code.BUYINFO :
					p = Store.newBuyInfo();
					decodeBuyInfoInto(in, (BuyInfo) p);
					break;
				case Code.RESREQTICKETDATA :
					p = Store.newResReqTicketData();
					decodeResReqTicketDataInto(in, (ResReqTicketData) p);
					break;
				case Code.CLIENTHELLOMSG :
					p = Store.newClientHelloMsg();
					decodeClientHelloMsgInto(in, (ClientHelloMsg) p);
					break;
				case Code.SESSIONKEYMSG :
					p = Store.newSessionKeyMsg();
					decodeSessionKeyMsgInto(in, (SessionKeyMsg) p);
					break;
				case Code.SERVERHELLOMSG :
					p = Store.newServerHelloMsg();
					decodeServerHelloMsgInto(in, (ServerHelloMsg) p);
					break;
				case Code.RECOVERYINFO :
					p = Store.newRecoveryInfo();
					decodeRecoveryInfoInto(in, (RecoveryInfo) p);
					break;
				case Code.RECOVERYDATA :
					p = Store.newRecoveryData();
					decodeRecoveryDataInto(in, (RecoveryData) p);
					break;
				case Code.TICKETRECEIVEDDATA :
					p = Store.newTicketReceivedData();
					decodeTicketReceivedDataInto(in, (TicketReceivedData) p);
					break;
				case Code.TICKETRECEIVEDDATARECOVERY :
					p = Store.newTicketReceivedDataRecovery();
					decodeTicketReceivedDataRecoveryInto(in,
							(TicketReceivedDataRecovery) p);
					break;
				case Code.CARDAUTHRESMSG :
					p = Store.newCardAuthResMsg();
					decodeCardAuthResMsgInto(in, (CardAuthResMsg) p);
					break;
				case Code.STAMPTICKETMSG :
					p = Store.newStampTicketMsg();
					decodeStampTicketMsgInto(in, (StampTicketMsg) p);
					break;
				case Code.RETURNALLITICKETSMSG :
					p = Store.newReturnAllITicketsMsg();
					decodeReturnAllITicketsMsgInto(in, (ReturnAllITicketsMsg) p);
					break;
				case Code.TICKET :
					p = Store.newTicket();
					decodeTicketInto(in, (Ticket) p);
					break;

			}
		} catch (ArrayIndexOutOfBoundsException e) {

			SimpleComm.raiseError();
		}

		if (p == null) {

			SimpleComm.raiseError();
		}
		return p;
	}

	private void encodeBuyTicket(BuyTicket c) {
		encodeDestination[encoding_length++] = Code.BUYTICKET;
		encode(c.ticketInfo);
		encode(c.pin);
	}

	private void decodeBuyTicketInto(byte[] in, BuyTicket into) {
		if (in[encoding_length++] != Code.BUYTICKET)
			SimpleComm.raiseError();
		decodeTicketInfoInto(in, (TicketInfo) (into.ticketInfo));
		decodeSecretInto(in, (Secret) (into.pin));
	}

	private void compSizeBuyTicket(BuyTicket c) {
		encoding_length++;
		compSize(c.ticketInfo);
		compSize(c.pin);
	}

	private void encodeReqTicket(ReqTicket c) {
		encodeDestination[encoding_length++] = Code.REQTICKET;
		encode(c.msg);
	}

	private void decodeReqTicketInto(byte[] in, ReqTicket into) {
		if (in[encoding_length++] != Code.REQTICKET)
			SimpleComm.raiseError();
		decodeEncDataSymmInto(in, (EncDataSymm) (into.msg));
	}

	private void compSizeReqTicket(ReqTicket c) {
		encoding_length++;
		compSize(c.msg);
	}

	private void encodeResReqTicket(ResReqTicket c) {
		encodeDestination[encoding_length++] = Code.RESREQTICKET;
		encode(c.msg);
	}

	private void decodeResReqTicketInto(byte[] in, ResReqTicket into) {
		if (in[encoding_length++] != Code.RESREQTICKET)
			SimpleComm.raiseError();
		decodeEncDataSymmInto(in, (EncDataSymm) (into.msg));
	}

	private void compSizeResReqTicket(ResReqTicket c) {
		encoding_length++;
		compSize(c.msg);
	}

	private void encodeDeleteTicket(DeleteTicket c) {
		encodeDestination[encoding_length++] = Code.DELETETICKET;
		encode(c.pin);
		encode(c.ticketID);
	}

	private void decodeDeleteTicketInto(byte[] in, DeleteTicket into) {
		if (in[encoding_length++] != Code.DELETETICKET)
			SimpleComm.raiseError();
		decodeSecretInto(in, (Secret) (into.pin));
		decodeIDInto(in, (ID) (into.ticketID));
	}

	private void compSizeDeleteTicket(DeleteTicket c) {
		encoding_length++;
		compSize(c.pin);
		compSize(c.ticketID);
	}

	private void encodeClientHello(ClientHello c) {
		encodeDestination[encoding_length++] = Code.CLIENTHELLO;
		encode(c.msg);
		encode(c.certCard);
		encode(c.publicKeyCard);
	}

	private void decodeClientHelloInto(byte[] in, ClientHello into) {
		if (in[encoding_length++] != Code.CLIENTHELLO)
			SimpleComm.raiseError();
		decodeEncDataAsymmInto(in, (EncDataAsymm) (into.msg));
		decodeSignedDataInto(in, (SignedData) (into.certCard));
		decodePublicKeyInto(in, (PublicKey) (into.publicKeyCard));
	}

	private void compSizeClientHello(ClientHello c) {
		encoding_length++;
		compSize(c.msg);
		compSize(c.certCard);
		compSize(c.publicKeyCard);
	}

	private void encodeSessionKey(SessionKey c) {
		encodeDestination[encoding_length++] = Code.SESSIONKEY;
		encode(c.msg);
	}

	private void decodeSessionKeyInto(byte[] in, SessionKey into) {
		if (in[encoding_length++] != Code.SESSIONKEY)
			SimpleComm.raiseError();
		decodeEncDataAsymmInto(in, (EncDataAsymm) (into.msg));
	}

	private void compSizeSessionKey(SessionKey c) {
		encoding_length++;
		compSize(c.msg);
	}

	private void encodeServerHello(ServerHello c) {
		encodeDestination[encoding_length++] = Code.SERVERHELLO;
		encode(c.msg);
	}

	private void decodeServerHelloInto(byte[] in, ServerHello into) {
		if (in[encoding_length++] != Code.SERVERHELLO)
			SimpleComm.raiseError();
		decodeEncDataAsymmInto(in, (EncDataAsymm) (into.msg));
	}

	private void compSizeServerHello(ServerHello c) {
		encoding_length++;
		compSize(c.msg);
	}

	private void encodeHandshake(Handshake c) {
		encodeDestination[encoding_length++] = Code.HANDSHAKE;
	}

	private void decodeHandshakeInto(byte[] in, Handshake into) {
		if (in[encoding_length++] != Code.HANDSHAKE)
			SimpleComm.raiseError();
	}

	private void compSizeHandshake(Handshake c) {
		encoding_length++;
	}

	private void encodeRecoverTicket(RecoverTicket c) {
		encodeDestination[encoding_length++] = Code.RECOVERTICKET;
		encode(c.pin);
	}

	private void decodeRecoverTicketInto(byte[] in, RecoverTicket into) {
		if (in[encoding_length++] != Code.RECOVERTICKET)
			SimpleComm.raiseError();
		decodeSecretInto(in, (Secret) (into.pin));
	}

	private void compSizeRecoverTicket(RecoverTicket c) {
		encoding_length++;
		compSize(c.pin);
	}

	private void encodeReqRecovery(ReqRecovery c) {
		encodeDestination[encoding_length++] = Code.REQRECOVERY;
		encode(c.recoveryInfo);
	}

	private void decodeReqRecoveryInto(byte[] in, ReqRecovery into) {
		if (in[encoding_length++] != Code.REQRECOVERY)
			SimpleComm.raiseError();
		decodeEncDataSymmInto(in, (EncDataSymm) (into.recoveryInfo));
	}

	private void compSizeReqRecovery(ReqRecovery c) {
		encoding_length++;
		compSize(c.recoveryInfo);
	}

	private void encodeResReqRecovery(ResReqRecovery c) {
		encodeDestination[encoding_length++] = Code.RESREQRECOVERY;
		encode(c.recoveryData);
	}

	private void decodeResReqRecoveryInto(byte[] in, ResReqRecovery into) {
		if (in[encoding_length++] != Code.RESREQRECOVERY)
			SimpleComm.raiseError();
		decodeEncDataSymmInto(in, (EncDataSymm) (into.recoveryData));
	}

	private void compSizeResReqRecovery(ResReqRecovery c) {
		encoding_length++;
		compSize(c.recoveryData);
	}

	private void encodeShowAllTickets(ShowAllTickets c) {
		encodeDestination[encoding_length++] = Code.SHOWALLTICKETS;
	}

	private void decodeShowAllTicketsInto(byte[] in, ShowAllTickets into) {
		if (in[encoding_length++] != Code.SHOWALLTICKETS)
			SimpleComm.raiseError();
	}

	private void compSizeShowAllTickets(ShowAllTickets c) {
		encoding_length++;
	}

	private void encodeReturnAllTickets(ReturnAllTickets c) {
		encodeDestination[encoding_length++] = Code.RETURNALLTICKETS;
		encode(c.tickets);
	}

	private void decodeReturnAllTicketsInto(byte[] in, ReturnAllTickets into) {
		if (in[encoding_length++] != Code.RETURNALLTICKETS)
			SimpleComm.raiseError();
		decodeListOfTicketInto(in, (ListOfTicket) (into.tickets));
	}

	private void compSizeReturnAllTickets(ReturnAllTickets c) {
		encoding_length++;
		compSize(c.tickets);
	}

	private void encodeTicketReceived(TicketReceived c) {
		encodeDestination[encoding_length++] = Code.TICKETRECEIVED;
		encode(c.msg);
	}

	private void decodeTicketReceivedInto(byte[] in, TicketReceived into) {
		if (in[encoding_length++] != Code.TICKETRECEIVED)
			SimpleComm.raiseError();
		decodeEncDataSymmInto(in, (EncDataSymm) (into.msg));
	}

	private void compSizeTicketReceived(TicketReceived c) {
		encoding_length++;
		compSize(c.msg);
	}

	private void encodeTicketReceivedRecovery(TicketReceivedRecovery c) {
		encodeDestination[encoding_length++] = Code.TICKETRECEIVEDRECOVERY;
		encode(c.msg);
	}

	private void decodeTicketReceivedRecoveryInto(byte[] in,
			TicketReceivedRecovery into) {
		if (in[encoding_length++] != Code.TICKETRECEIVEDRECOVERY)
			SimpleComm.raiseError();
		decodeEncDataSymmInto(in, (EncDataSymm) (into.msg));
	}

	private void compSizeTicketReceivedRecovery(TicketReceivedRecovery c) {
		encoding_length++;
		compSize(c.msg);
	}

	private void encodeCardAuth(CardAuth c) {
		encodeDestination[encoding_length++] = Code.CARDAUTH;
		encode(c.publicKeyTerminal);
	}

	private void decodeCardAuthInto(byte[] in, CardAuth into) {
		if (in[encoding_length++] != Code.CARDAUTH)
			SimpleComm.raiseError();
		decodePublicKeyInto(in, (PublicKey) (into.publicKeyTerminal));
	}

	private void compSizeCardAuth(CardAuth c) {
		encoding_length++;
		compSize(c.publicKeyTerminal);
	}

	private void encodeCardAuthRes(CardAuthRes c) {
		encodeDestination[encoding_length++] = Code.CARDAUTHRES;
		encode(c.publicKeyCard);
		encode(c.certCard);
		encode(c.msg);
	}

	private void decodeCardAuthResInto(byte[] in, CardAuthRes into) {
		if (in[encoding_length++] != Code.CARDAUTHRES)
			SimpleComm.raiseError();
		decodePublicKeyInto(in, (PublicKey) (into.publicKeyCard));
		decodeSignedDataInto(in, (SignedData) (into.certCard));
		decodeEncDataAsymmInto(in, (EncDataAsymm) (into.msg));
	}

	private void compSizeCardAuthRes(CardAuthRes c) {
		encoding_length++;
		compSize(c.publicKeyCard);
		compSize(c.certCard);
		compSize(c.msg);
	}

	private void encodeStampTicket(StampTicket c) {
		encodeDestination[encoding_length++] = Code.STAMPTICKET;
		encode(c.msg);
	}

	private void decodeStampTicketInto(byte[] in, StampTicket into) {
		if (in[encoding_length++] != Code.STAMPTICKET)
			SimpleComm.raiseError();
		decodeEncDataAsymmInto(in, (EncDataAsymm) (into.msg));
	}

	private void compSizeStampTicket(StampTicket c) {
		encoding_length++;
		compSize(c.msg);
	}

	private void encodeStampTicketSuccessful(StampTicketSuccessful c) {
		encodeDestination[encoding_length++] = Code.STAMPTICKETSUCCESSFUL;
	}

	private void decodeStampTicketSuccessfulInto(byte[] in,
			StampTicketSuccessful into) {
		if (in[encoding_length++] != Code.STAMPTICKETSUCCESSFUL)
			SimpleComm.raiseError();
	}

	private void compSizeStampTicketSuccessful(StampTicketSuccessful c) {
		encoding_length++;
	}

	private void encodeReturnAllITickets(ReturnAllITickets c) {
		encodeDestination[encoding_length++] = Code.RETURNALLITICKETS;
		encode(c.msg);
	}

	private void decodeReturnAllITicketsInto(byte[] in, ReturnAllITickets into) {
		if (in[encoding_length++] != Code.RETURNALLITICKETS)
			SimpleComm.raiseError();
		decodeEncDataSymmInto(in, (EncDataSymm) (into.msg));
	}

	private void compSizeReturnAllITickets(ReturnAllITickets c) {
		encoding_length++;
		compSize(c.msg);
	}

	private void encodeShowAllITickets(ShowAllITickets c) {
		encodeDestination[encoding_length++] = Code.SHOWALLITICKETS;
	}

	private void decodeShowAllITicketsInto(byte[] in, ShowAllITickets into) {
		if (in[encoding_length++] != Code.SHOWALLITICKETS)
			SimpleComm.raiseError();
	}

	private void compSizeShowAllITickets(ShowAllITickets c) {
		encoding_length++;
	}

	private void encodeTicketInfo(TicketInfo c) {
		encodeDestination[encoding_length++] = Code.TICKETINFO;
		encodeByteArray((byte[]) c.departure);
		encodeByteArray((byte[]) c.destination);
		encodeShort((short) c.trainClass);
		encodeByteArray((byte[]) c.expiration);
	}

	private void decodeTicketInfoInto(byte[] in, TicketInfo into) {
		if (in[encoding_length++] != Code.TICKETINFO)
			SimpleComm.raiseError();
		decodeByteArrayInto(in, into.departure);
		decodeByteArrayInto(in, into.destination);
		into.trainClass = decodeShort(in);
		decodeByteArrayInto(in, into.expiration);
	}

	private void compSizeTicketInfo(TicketInfo c) {
		encoding_length++;
		compSizeByteArray(c.departure);
		compSizeByteArray(c.destination);
		compSizeShort(c.trainClass);
		compSizeByteArray(c.expiration);
	}

	private void encodeSecret(Secret c) {
		encodeDestination[encoding_length++] = Code.SECRET;
		encodeByteArray((byte[]) c.secret);
	}

	private void decodeSecretInto(byte[] in, Secret into) {
		if (in[encoding_length++] != Code.SECRET)
			SimpleComm.raiseError();
		decodeByteArrayInto(in, into.secret);
	}

	private void compSizeSecret(Secret c) {
		encoding_length++;
		compSizeByteArray(c.secret);
	}

	private void encodeEncDataSymm(EncDataSymm c) {

		encodeDestination[encoding_length++] = Code.ENCDATASYMM;
		encodeByteArray(c.encrypted);

		encodeShort(c.plainlength);
		encodeShort(c.enclength);

		encodeByte(c.plainDataType);
	}

	private void decodeEncDataSymmInto(byte[] in, EncDataSymm into) {
		if (in[encoding_length++] != Code.ENCDATASYMM)
			SimpleComm.raiseError();
		//public byte[] encrypted;
		//public short plainlength;
		//public short enclength;
		//public byte plainDataType;
		decodeByteArrayInto(in, into.encrypted);
		into.plainlength = decodeShort(in);
		into.enclength = decodeShort(in);
		byte b = decodeByte(in);
		into.plainDataType = b;
		if (b != Code.BUYINFO && b != Code.RESREQTICKETDATA
				&& b != Code.CLIENTHELLOMSG && b != Code.SESSIONKEYMSG
				&& b != Code.SERVERHELLOMSG && b != Code.RECOVERYINFO
				&& b != Code.RECOVERYDATA && b != Code.TICKETRECEIVEDDATA
				&& b != Code.TICKETRECEIVEDDATARECOVERY
				&& b != Code.CARDAUTHRESMSG && b != Code.STAMPTICKETMSG
				&& b != Code.RETURNALLITICKETSMSG && b != Code.TICKET)
			SimpleComm.raiseError();
	}

	private void compSizeEncDataSymm(EncDataSymm c) {
		encoding_length++;
		compSizeByteArray(c.encrypted);
		encoding_length += 8; // short + short + byte
	}

	private void encodeBuyInfo(BuyInfo c) {
		encodeDestination[encoding_length++] = Code.BUYINFO;
		encode(c.ccInfo);
		encode(c.ticketInfo);
		encode(c.cardID);
		encode(c.transactionId);
	}

	private void decodeBuyInfoInto(byte[] in, BuyInfo into) {
		if (in[encoding_length++] != Code.BUYINFO)
			SimpleComm.raiseError();
		decodeCCInfoInto(in, (CCInfo) (into.ccInfo));
		decodeTicketInfoInto(in, (TicketInfo) (into.ticketInfo));
		decodeIDInto(in, (ID) (into.cardID));
		decodeNonceInto(in, (Nonce) (into.transactionId));
	}

	private void compSizeBuyInfo(BuyInfo c) {
		encoding_length++;
		compSize(c.ccInfo);
		compSize(c.ticketInfo);
		compSize(c.cardID);
		compSize(c.transactionId);
	}

	private void encodeResReqTicketData(ResReqTicketData c) {
		encodeDestination[encoding_length++] = Code.RESREQTICKETDATA;
		encode(c.ticket);
		encode(c.transactionId);
		encode(c.nonceDeleteTicket);
	}

	private void decodeResReqTicketDataInto(byte[] in, ResReqTicketData into) {
		if (in[encoding_length++] != Code.RESREQTICKETDATA)
			SimpleComm.raiseError();
		decodeTicketInto(in, (Ticket) (into.ticket));
		decodeNonceInto(in, (Nonce) (into.transactionId));
		decodeNonceInto(in, (Nonce) (into.nonceDeleteTicket));
	}

	private void compSizeResReqTicketData(ResReqTicketData c) {
		encoding_length++;
		compSize(c.ticket);
		compSize(c.transactionId);
		compSize(c.nonceDeleteTicket);
	}

	private void encodeID(ID c) {
		encodeDestination[encoding_length++] = Code.ID;
		encodeShort((short) c.id);
	}

	private void decodeIDInto(byte[] in, ID into) {
		if (in[encoding_length++] != Code.ID)
			SimpleComm.raiseError();
		into.id = decodeShort(in);
	}

	private void compSizeID(ID c) {
		encoding_length++;
		compSizeShort(c.id);
	}

	private void encodePublicKey(PublicKey c) {
		encodeDestination[encoding_length++] = Code.PUBLICKEY;

		encodeByteArray(c.modulus);
	}

	private void decodePublicKeyInto(byte[] in, PublicKey into) {
		if (in[encoding_length++] != Code.PUBLICKEY)
			SimpleComm.raiseError();
		decodeByteArrayInto(in, into.modulus);
		into.updateKey();
	}

	private void compSizePublicKey(PublicKey c) {
		encoding_length++;

		compSizeByteArray(c.modulus);

	}

	private void encodeSignedData(SignedData c) {
		encodeDestination[encoding_length++] = Code.SIGNEDDATA;
		encodeByteArray((byte[]) c.signed);
	}

	private void decodeSignedDataInto(byte[] in, SignedData into) {
		if (in[encoding_length++] != Code.SIGNEDDATA)
			SimpleComm.raiseError();
		decodeByteArrayInto(in, into.signed);
	}

	private void compSizeSignedData(SignedData c) {
		encoding_length++;
		compSizeByteArray(c.signed);
	}

	private void encodeEncDataAsymm(EncDataAsymm c) {

		encodeDestination[encoding_length++] = Code.ENCDATAASYMM;
		encodeByteArray(c.encrypted);

		encodeShort(c.plainlength);
		encodeShort(c.enclength);

		encodeByte(c.plainDataType);
	}

	private void decodeEncDataAsymmInto(byte[] in, EncDataAsymm into) {
		if (in[encoding_length++] != Code.ENCDATAASYMM)
			SimpleComm.raiseError();
		//public byte[] encrypted;
		//public short plainlength;
		//public short enclength;
		//public byte plainDataType;
		decodeByteArrayInto(in, into.encrypted);
		into.plainlength = decodeShort(in);
		into.enclength = decodeShort(in);
		byte b = decodeByte(in);
		into.plainDataType = b;
		if (b != Code.BUYINFO && b != Code.RESREQTICKETDATA
				&& b != Code.CLIENTHELLOMSG && b != Code.SESSIONKEYMSG
				&& b != Code.SERVERHELLOMSG && b != Code.RECOVERYINFO
				&& b != Code.RECOVERYDATA && b != Code.TICKETRECEIVEDDATA
				&& b != Code.TICKETRECEIVEDDATARECOVERY
				&& b != Code.CARDAUTHRESMSG && b != Code.STAMPTICKETMSG
				&& b != Code.RETURNALLITICKETSMSG && b != Code.TICKET)
			SimpleComm.raiseError();
	}

	private void compSizeEncDataAsymm(EncDataAsymm c) {
		encoding_length++;
		compSizeByteArray(c.encrypted);
		encoding_length += 8; // short + short + byte
	}

	private void encodeCertificate(Certificate c) {
		encodeDestination[encoding_length++] = Code.CERTIFICATE;
		encode(c.publicKey);
	}

	private void decodeCertificateInto(byte[] in, Certificate into) {
		if (in[encoding_length++] != Code.CERTIFICATE)
			SimpleComm.raiseError();
		decodePublicKeyInto(in, (PublicKey) (into.publicKey));
	}

	private void compSizeCertificate(Certificate c) {
		encoding_length++;
		compSize(c.publicKey);
	}

	private void encodeClientHelloMsg(ClientHelloMsg c) {
		encodeDestination[encoding_length++] = Code.CLIENTHELLOMSG;
		encode(c.nonceCard);
	}

	private void decodeClientHelloMsgInto(byte[] in, ClientHelloMsg into) {
		if (in[encoding_length++] != Code.CLIENTHELLOMSG)
			SimpleComm.raiseError();
		decodeNonceInto(in, (Nonce) (into.nonceCard));
	}

	private void compSizeClientHelloMsg(ClientHelloMsg c) {
		encoding_length++;
		compSize(c.nonceCard);
	}

	private void encodeSessionKeyMsg(SessionKeyMsg c) {
		encodeDestination[encoding_length++] = Code.SESSIONKEYMSG;
		encode(c.nonceService);
		encode(c.sessionKey);
	}

	private void decodeSessionKeyMsgInto(byte[] in, SessionKeyMsg into) {
		if (in[encoding_length++] != Code.SESSIONKEYMSG)
			SimpleComm.raiseError();
		decodeNonceInto(in, (Nonce) (into.nonceService));
		decodeSymmKeyInto(in, (SymmKey) (into.sessionKey));
	}

	private void compSizeSessionKeyMsg(SessionKeyMsg c) {
		encoding_length++;
		compSize(c.nonceService);
		compSize(c.sessionKey);
	}

	private void encodeServerHelloMsg(ServerHelloMsg c) {
		encodeDestination[encoding_length++] = Code.SERVERHELLOMSG;
		encode(c.nonceService);
		encode(c.nonceCard);
	}

	private void decodeServerHelloMsgInto(byte[] in, ServerHelloMsg into) {
		if (in[encoding_length++] != Code.SERVERHELLOMSG)
			SimpleComm.raiseError();
		decodeNonceInto(in, (Nonce) (into.nonceService));
		decodeNonceInto(in, (Nonce) (into.nonceCard));
	}

	private void compSizeServerHelloMsg(ServerHelloMsg c) {
		encoding_length++;
		compSize(c.nonceService);
		compSize(c.nonceCard);
	}

	private void encodeRecoveryInfo(RecoveryInfo c) {
		encodeDestination[encoding_length++] = Code.RECOVERYINFO;
		encode(c.cardId);
		encode(c.transactionId);
	}

	private void decodeRecoveryInfoInto(byte[] in, RecoveryInfo into) {
		if (in[encoding_length++] != Code.RECOVERYINFO)
			SimpleComm.raiseError();
		decodeIDInto(in, (ID) (into.cardId));
		decodeNonceInto(in, (Nonce) (into.transactionId));
	}

	private void compSizeRecoveryInfo(RecoveryInfo c) {
		encoding_length++;
		compSize(c.cardId);
		compSize(c.transactionId);
	}

	private void encodeRecoveryData(RecoveryData c) {
		encodeDestination[encoding_length++] = Code.RECOVERYDATA;
		encode(c.ticket);
		encode(c.transactionId);
		encode(c.nonceDeleteTicket);
	}

	private void decodeRecoveryDataInto(byte[] in, RecoveryData into) {
		if (in[encoding_length++] != Code.RECOVERYDATA)
			SimpleComm.raiseError();
		decodeTicketInto(in, (Ticket) (into.ticket));
		decodeNonceInto(in, (Nonce) (into.transactionId));
		decodeNonceInto(in, (Nonce) (into.nonceDeleteTicket));
	}

	private void compSizeRecoveryData(RecoveryData c) {
		encoding_length++;
		compSize(c.ticket);
		compSize(c.transactionId);
		compSize(c.nonceDeleteTicket);
	}

	private void encodeListOfTicket(ListOfTicket c) {
		encodeDestination[encoding_length++] = Code.LISTOFTICKET;
		encodeDestination[encoding_length++] = (byte) c.index();
		for (int i = 0; i < c.index(); i++)
			encode(c.elems[i]);
	}

	private void decodeListOfTicketInto(byte[] in, ListOfTicket into) {
		if (in[encoding_length++] != Code.LISTOFTICKET)
			SimpleComm.raiseError();
		byte index = in[encoding_length++];
		if (index > into.size())
			SimpleComm.raiseError();
		into.index = index;
		for (int i = 0; i < index; i++)
			decodeTicketInto(in, into.elems[i]);
	}

	private void compSizeListOfTicket(ListOfTicket c) {
		encoding_length++;
		encoding_length++;
		for (int i = 0; i < c.index(); i++)
			compSize(c.elems[i]);
	}

	private void encodeTicketReceivedData(TicketReceivedData c) {
		encodeDestination[encoding_length++] = Code.TICKETRECEIVEDDATA;
		encode(c.cardId);
		encode(c.nonceDeleteTicket);
	}

	private void decodeTicketReceivedDataInto(byte[] in, TicketReceivedData into) {
		if (in[encoding_length++] != Code.TICKETRECEIVEDDATA)
			SimpleComm.raiseError();
		decodeIDInto(in, (ID) (into.cardId));
		decodeNonceInto(in, (Nonce) (into.nonceDeleteTicket));
	}

	private void compSizeTicketReceivedData(TicketReceivedData c) {
		encoding_length++;
		compSize(c.cardId);
		compSize(c.nonceDeleteTicket);
	}

	private void encodeTicketReceivedDataRecovery(TicketReceivedDataRecovery c) {
		encodeDestination[encoding_length++] = Code.TICKETRECEIVEDDATARECOVERY;
		encode(c.cardId);
		encode(c.nonceDeleteTicket);
	}

	private void decodeTicketReceivedDataRecoveryInto(byte[] in,
			TicketReceivedDataRecovery into) {
		if (in[encoding_length++] != Code.TICKETRECEIVEDDATARECOVERY)
			SimpleComm.raiseError();
		decodeIDInto(in, (ID) (into.cardId));
		decodeNonceInto(in, (Nonce) (into.nonceDeleteTicket));
	}

	private void compSizeTicketReceivedDataRecovery(TicketReceivedDataRecovery c) {
		encoding_length++;
		compSize(c.cardId);
		compSize(c.nonceDeleteTicket);
	}

	private void encodeCardAuthResMsg(CardAuthResMsg c) {
		encodeDestination[encoding_length++] = Code.CARDAUTHRESMSG;
		encode(c.sessionKey);
	}

	private void decodeCardAuthResMsgInto(byte[] in, CardAuthResMsg into) {
		if (in[encoding_length++] != Code.CARDAUTHRESMSG)
			SimpleComm.raiseError();
		decodeSymmKeyInto(in, (SymmKey) (into.sessionKey));
	}

	private void compSizeCardAuthResMsg(CardAuthResMsg c) {
		encoding_length++;
		compSize(c.sessionKey);
	}

	private void encodeStampTicketMsg(StampTicketMsg c) {
		encodeDestination[encoding_length++] = Code.STAMPTICKETMSG;
		encode(c.ticketId);
		encode(c.pin);
		encode(c.stamp);
	}

	private void decodeStampTicketMsgInto(byte[] in, StampTicketMsg into) {
		if (in[encoding_length++] != Code.STAMPTICKETMSG)
			SimpleComm.raiseError();
		decodeIDInto(in, (ID) (into.ticketId));
		decodeSecretInto(in, (Secret) (into.pin));
		decodeStampInto(in, (Stamp) (into.stamp));
	}

	private void compSizeStampTicketMsg(StampTicketMsg c) {
		encoding_length++;
		compSize(c.ticketId);
		compSize(c.pin);
		compSize(c.stamp);
	}

	private void encodeReturnAllITicketsMsg(ReturnAllITicketsMsg c) {
		encodeDestination[encoding_length++] = Code.RETURNALLITICKETSMSG;
		encode(c.tickets);
	}

	private void decodeReturnAllITicketsMsgInto(byte[] in,
			ReturnAllITicketsMsg into) {
		if (in[encoding_length++] != Code.RETURNALLITICKETSMSG)
			SimpleComm.raiseError();
		decodeListOfTicketInto(in, (ListOfTicket) (into.tickets));
	}

	private void compSizeReturnAllITicketsMsg(ReturnAllITicketsMsg c) {
		encoding_length++;
		compSize(c.tickets);
	}

	private void encodeCCInfo(CCInfo c) {
		encodeDestination[encoding_length++] = Code.CCINFO;
		encodeByteArray((byte[]) c.CCHolderName);
		encodeByteArray((byte[]) c.CCSecNumber);
		encodeByteArray((byte[]) c.CCNumber);
		encodeByteArray((byte[]) c.CCType);
		encodeByteArray((byte[]) c.CCExpDate);
	}

	private void decodeCCInfoInto(byte[] in, CCInfo into) {
		if (in[encoding_length++] != Code.CCINFO)
			SimpleComm.raiseError();
		decodeByteArrayInto(in, into.CCHolderName);
		decodeByteArrayInto(in, into.CCSecNumber);
		decodeByteArrayInto(in, into.CCNumber);
		decodeByteArrayInto(in, into.CCType);
		decodeByteArrayInto(in, into.CCExpDate);
	}

	private void compSizeCCInfo(CCInfo c) {
		encoding_length++;
		compSizeByteArray(c.CCHolderName);
		compSizeByteArray(c.CCSecNumber);
		compSizeByteArray(c.CCNumber);
		compSizeByteArray(c.CCType);
		compSizeByteArray(c.CCExpDate);
	}

	private void encodeNonce(Nonce c) {
		encodeDestination[encoding_length++] = Code.NONCE;
		encodeByteArray((byte[]) c.nonce);
	}

	private void decodeNonceInto(byte[] in, Nonce into) {
		if (in[encoding_length++] != Code.NONCE)
			SimpleComm.raiseError();
		decodeByteArrayInto(in, into.nonce);
	}

	private void compSizeNonce(Nonce c) {
		encoding_length++;
		compSizeByteArray(c.nonce);
	}

	private void encodeTicket(Ticket c) {
		encodeDestination[encoding_length++] = Code.TICKET;
		encode(c.id);
		encode(c.ticketInfo);
		encode(c.stamp);
	}

	private void decodeTicketInto(byte[] in, Ticket into) {
		if (in[encoding_length++] != Code.TICKET)
			SimpleComm.raiseError();
		decodeIDInto(in, (ID) (into.id));
		decodeTicketInfoInto(in, (TicketInfo) (into.ticketInfo));
		decodeStampInto(in, (Stamp) (into.stamp));
	}

	private void compSizeTicket(Ticket c) {
		encoding_length++;
		compSize(c.id);
		compSize(c.ticketInfo);
		compSize(c.stamp);
	}

	private void encodeSymmKey(SymmKey c) {
		encodeDestination[encoding_length++] = Code.SYMMKEY;
		// public byte[] key; - ok
		// private DESKey apikey; - ignored
		encodeByteArray(c.key);
	}

	private void decodeSymmKeyInto(byte[] in, SymmKey into) {
		if (in[encoding_length++] != Code.SYMMKEY)
			SimpleComm.raiseError();
		// public byte[] key; - ok
		// private DESKey apikey; - handled by updateKey() below
		decodeByteArrayInto(in, into.key);
		into.updateKey();
	}

	private void compSizeSymmKey(SymmKey c) {
		encoding_length++;
		// public byte[] key; - ok
		// private DESKey apikey; - ignored
		compSizeByteArray(c.key);
	}

	private void encodeStamp(Stamp c) {
		encodeDestination[encoding_length++] = Code.STAMP;
		encodeBoolean((boolean) c.stamped);
		encode(c.stampTime);
	}

	private void decodeStampInto(byte[] in, Stamp into) {
		if (in[encoding_length++] != Code.STAMP)
			SimpleComm.raiseError();
		into.stamped = decodeBoolean(in);
		decodeDateTimeInto(in, (DateTime) (into.stampTime));
	}

	private void compSizeStamp(Stamp c) {
		encoding_length++;
		compSizeBoolean(c.stamped);
		compSize(c.stampTime);
	}

	private void encodeDateTime(DateTime c) {
		encodeDestination[encoding_length++] = Code.DATETIME;
		encodeByteArray((byte[]) c.dateTime);
	}

	private void decodeDateTimeInto(byte[] in, DateTime into) {
		if (in[encoding_length++] != Code.DATETIME)
			SimpleComm.raiseError();
		decodeByteArrayInto(in, into.dateTime);
	}

	private void compSizeDateTime(DateTime c) {
		encoding_length++;
		compSizeByteArray(c.dateTime);
	}

	public void encodePrivateKey(PrivateKey c) {
		encodeDestination[encoding_length++] = Code.PRIVATEKEY;

		encodeByteArray(c.modulus);
		encodeByteArray(c.privexponent);

	}

	private void decodePrivateKeyInto(byte[] in, PrivateKey into) {
		if (in[encoding_length++] != Code.PRIVATEKEY)
			SimpleComm.raiseError();
		decodeByteArrayInto(in, into.modulus);
		decodeByteArrayInto(in, into.privexponent);
		into.updateKey();
	}

	private void compSizePrivateKey(PrivateKey c) {
		encoding_length++;

		compSizeByteArray(c.modulus);
		compSizeByteArray(c.privexponent);

	}

	public void decodeETicketCard(byte[] in, short offset,
			ETicketCard thisETicketCard) {
		encoding_length = offset;
		if (in[encoding_length++] != Code.ETICKETCARD)
			SimpleComm.raiseError();

		if (in[encoding_length] != Code.IGNORE) {
			decodeCCInfoInto(in, thisETicketCard.ccInfo);
		} else
			encoding_length++;

		if (in[encoding_length] != Code.IGNORE) {
			decodeSignedDataInto(in, thisETicketCard.certCard);
		} else
			encoding_length++;

		if (in[encoding_length] != Code.IGNORE) {
			decodeIDInto(in, thisETicketCard.id);
		} else
			encoding_length++;

		if (in[encoding_length] != Code.IGNORE) {
			decodeSecretInto(in, thisETicketCard.pin);
		} else
			encoding_length++;

		if (in[encoding_length] != Code.IGNORE) {
			decodePrivateKeyInto(in, thisETicketCard.privateKeyCard);
		} else
			encoding_length++;

		if (in[encoding_length] != Code.IGNORE) {
			decodePublicKeyInto(in, thisETicketCard.publicKeyCard);
		} else
			encoding_length++;

		if (in[encoding_length] != Code.IGNORE) {
			decodePublicKeyInto(in, thisETicketCard.publicKeyServer);
		} else
			encoding_length++;

		if (in[encoding_length] != Code.IGNORE) {
			thisETicketCard.state = decodeByte(in);
		} else
			encoding_length++;
	}

}
