package generated.eTicket.javacard.eTicketCard;

import javacard.framework.*;

/**
 * Generated Smartcard for eTicket model
 */
public class ETicketCard extends SimpleComm {

	public PrivateKey privateKeyCard;
	public PublicKey publicKeyCard;
	public PublicKey publicKeyServer;
	public ID id;
	public Secret pin;
	public SymmKey sessionKey;
	public Nonce nonceCard;
	public Nonce transactionId;
	public boolean waitingForTicket = false;
	public PublicKey publicKeyTerminal;
	public CCInfo ccInfo;
	public byte state;
	public ListOfTicket tickets;
	public SignedData certCard;

	/**
	 * Constructor
	 */
	public ETicketCard() {
		privateKeyCard = new PrivateKey();
		publicKeyCard = new PublicKey();
		publicKeyServer = new PublicKey();
		id = new ID();
		pin = new Secret();
		sessionKey = new SymmKey();
		nonceCard = new Nonce();
		transactionId = new Nonce();
		waitingForTicket = false;
		publicKeyTerminal = new PublicKey();
		ccInfo = new CCInfo();
		state = StateCard.APPLET_UNINITIALIZED;
		tickets = new ListOfTicket((short) 10);
		certCard = new SignedData();
	}

	public void appletInitialization(byte[] buf, short offset) {
		if (state != StateCard.APPLET_UNINITIALIZED)
			stop();
		Coding.getInstance().decodeETicketCard(buf, offset, this);
	}

	/**
	 * This init method is used for testcases to quickly initialize a card
	 */
	public ETicketCard init(CCInfo ccInfo, SignedData certCard, ID id,
			Secret pin, PrivateKey privateKeyCard, PublicKey publicKeyCard,
			PublicKey publicKeyServer, byte state) {

		this.ccInfo = ccInfo;
		this.certCard = certCard;
		this.id = id;
		this.pin = pin;
		this.privateKeyCard = privateKeyCard;
		this.publicKeyCard = publicKeyCard;
		this.publicKeyServer = publicKeyServer;
		this.state = state;

		sessionKey = new SymmKey();
		nonceCard = new Nonce();
		transactionId = new Nonce();
		waitingForTicket = false;
		publicKeyTerminal = new PublicKey();
		tickets = new ListOfTicket((short) 10);

		return this;
	}

	public ETicketCard init() {
		privateKeyCard = new PrivateKey();
		publicKeyCard = new PublicKey();
		publicKeyServer = new PublicKey();
		id = new ID();
		pin = new Secret();
		sessionKey = new SymmKey();
		nonceCard = new Nonce();
		transactionId = new Nonce();
		waitingForTicket = false;
		publicKeyTerminal = new PublicKey();
		ccInfo = new CCInfo();
		state = StateCard.IDLE;
		tickets = new ListOfTicket((short) 10);
		certCard = new SignedData();

		return this;
	}

	/**
	 * process-method, is called from SimpleComm when a message arrives
	 * 
	 * @param inmsg message to process
	 */
	public void process(Message inmsg) {
		if (state == StateCard.APPLET_UNINITIALIZED)
			stop();

		currentGate.callConstraints();

		switch (inmsg.getCode()) {
			case Code.BUYTICKET :
				processBuyTicket((BuyTicket) inmsg);
				break;
			case Code.RESREQTICKET :
				processResReqTicket((ResReqTicket) inmsg);
				break;
			case Code.DELETETICKET :
				processDeleteTicket((DeleteTicket) inmsg);
				break;
			case Code.HANDSHAKE :
				processHandshake((Handshake) inmsg);
				break;
			case Code.SERVERHELLO :
				processServerHello((ServerHello) inmsg);
				break;
			case Code.RECOVERTICKET :
				processRecoverTicket((RecoverTicket) inmsg);
				break;
			case Code.RESREQRECOVERY :
				processResReqRecovery((ResReqRecovery) inmsg);
				break;
			case Code.SHOWALLTICKETS :
				processShowAllTickets((ShowAllTickets) inmsg);
				break;
			case Code.CARDAUTH :
				processCardAuth((CardAuth) inmsg);
				break;
			case Code.STAMPTICKET :
				processStampTicket((StampTicket) inmsg);
				break;
			case Code.SHOWALLITICKETS :
				processShowAllITickets((ShowAllITickets) inmsg);
				break;
			//Unknown Message
			default :
				stop();
		}

		currentGate.callConstraints();
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	public void processBuyTicket(BuyTicket inmsg) {
		TicketInfo ticketInfo = inmsg.ticketInfo;
		Secret pin = inmsg.pin;
		if ((((state == StateCard.AUTHENTICATED && pin.equals(this.pin)) && waitingForTicket == false) && tickets
				.hasFree())) {
			waitingForTicket = true;
			transactionId.copy(Nonce.generateNonce());
			BuyInfo msg = Store.newBuyInfo(ccInfo, ticketInfo, id,
					transactionId);
			EncDataSymm enc = EncDataSymm.encrypt(sessionKey, msg);
			state = StateCard.EXP_TICKET;
			sendMsg(Store.newReqTicket(enc));
		} else {
			state = StateCard.IDLE;
			stop();
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	public void processResReqTicket(ResReqTicket inmsg) {
		EncDataSymm encResponse = inmsg.msg;
		if (state == StateCard.EXP_TICKET) {
			ResReqTicketData msg = (ResReqTicketData) (EncDataSymm.decrypt(
					sessionKey, encResponse));
			if (transactionId.equals(msg.transactionId)) {
				tickets.set(msg.ticket);
				waitingForTicket = false;
				TicketReceivedData trd = Store.newTicketReceivedData(id,
						msg.nonceDeleteTicket);
				EncDataSymm enc = EncDataSymm.encrypt(sessionKey, trd);
				state = StateCard.IDLE;
				sendMsg(Store.newTicketReceived(enc));
			} else {
				state = StateCard.IDLE;
				stop();
			}
		} else {
			state = StateCard.IDLE;
			stop();
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	public void processDeleteTicket(DeleteTicket inmsg) {
		Secret pin = inmsg.pin;
		ID ticketID = inmsg.ticketID;
		if ((state == StateCard.IDLE && this.pin.equals(pin))) {
			tickets.remove(ticketID);
		} else {
			state = StateCard.IDLE;
			stop();
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	public void processHandshake(Handshake inmsg) {

		if (state == StateCard.IDLE) {
			nonceCard.copy(Nonce.generateNonce());
			ClientHelloMsg chm = Store.newClientHelloMsg(nonceCard);
			EncDataAsymm enc = EncDataAsymm.encrypt(publicKeyServer, chm);
			state = StateCard.WAIT_FOR_SERVER_HELLO;
			sendMsg(Store.newClientHello(enc, certCard, publicKeyCard));
		} else {
			state = StateCard.IDLE;
			stop();
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	public void processServerHello(ServerHello inmsg) {
		EncDataAsymm enc1 = inmsg.msg;
		if (state == StateCard.WAIT_FOR_SERVER_HELLO) {
			ServerHelloMsg shm = (ServerHelloMsg) (EncDataAsymm.decrypt(
					privateKeyCard, enc1));
			if (shm.nonceCard.equals(nonceCard)) {
				sessionKey.copy(SymmKey.generateKey());
				SessionKeyMsg skm = Store.newSessionKeyMsg(shm.nonceService,
						sessionKey);
				EncDataAsymm enc = EncDataAsymm.encrypt(publicKeyServer, skm);
				state = StateCard.AUTHENTICATED;
				sendMsg(Store.newSessionKey(enc));
			} else {
				state = StateCard.IDLE;
				stop();
			}
		} else {
			state = StateCard.IDLE;
			stop();
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	public void processRecoverTicket(RecoverTicket inmsg) {
		Secret pin = inmsg.pin;
		if (((waitingForTicket && state == StateCard.AUTHENTICATED) && pin
				.equals(this.pin))) {
			RecoveryInfo recoveryInfo = Store
					.newRecoveryInfo(id, transactionId);
			EncDataSymm enc = EncDataSymm.encrypt(sessionKey, recoveryInfo);
			state = StateCard.EXP_REC_TICKET;
			sendMsg(Store.newReqRecovery(enc));
		} else {
			state = StateCard.IDLE;
			stop();
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	public void processResReqRecovery(ResReqRecovery inmsg) {
		EncDataSymm resEnc = inmsg.recoveryData;
		if (state == StateCard.EXP_REC_TICKET) {
			RecoveryData msg = (RecoveryData) (EncDataSymm.decrypt(sessionKey,
					resEnc));
			if (transactionId.equals(msg.transactionId)) {
				tickets.set(msg.ticket);
				TicketReceivedDataRecovery trd = Store
						.newTicketReceivedDataRecovery(id,
								msg.nonceDeleteTicket);
				EncDataSymm enc = EncDataSymm.encrypt(sessionKey, trd);
				waitingForTicket = false;
				sendMsg(Store.newTicketReceivedRecovery(enc));
			} else {
				state = StateCard.IDLE;
				stop();
			}
		} else {
			state = StateCard.IDLE;
			stop();
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	public void processShowAllTickets(ShowAllTickets inmsg) {

		if (state == StateCard.IDLE) {
			sendMsg(Store.newReturnAllTickets(tickets));
		} else {
			state = StateCard.IDLE;
			stop();
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	public void processCardAuth(CardAuth inmsg) {
		PublicKey publicKeyTerminal = inmsg.publicKeyTerminal;
		if (state == StateCard.IDLE) {
			this.publicKeyTerminal.copy(publicKeyTerminal);
			sessionKey.copy(SymmKey.generateKey());
			CardAuthResMsg msg = Store.newCardAuthResMsg(sessionKey);
			EncDataAsymm enc = EncDataAsymm.encrypt(publicKeyTerminal, msg);
			sendMsg(Store.newCardAuthRes(publicKeyCard, certCard, enc));
		} else {
			state = StateCard.IDLE;
			stop();
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	public void processStampTicket(StampTicket inmsg) {
		EncDataAsymm enc = inmsg.msg;
		if (state == StateCard.IDLE) {
			StampTicketMsg std = (StampTicketMsg) (EncDataAsymm.decrypt(
					privateKeyCard, enc));
			if (pin.equals(std.pin)) {
				Ticket ticket = tickets.get(std.ticketId);
				if (ticket.stamp.stamped) {
					state = StateCard.IDLE;
					stop();
				} else {
					ticket.stamp.copy(std.stamp);
					ticket.stamp.stamped = true;
					tickets.set(ticket);
					sendMsg(Store.newStampTicketSuccessful());
				}
			} else {
				state = StateCard.IDLE;
				stop();
			}
		} else {
			state = StateCard.IDLE;
			stop();
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	public void processShowAllITickets(ShowAllITickets inmsg) {

		if (state == StateCard.IDLE) {
			ReturnAllITicketsMsg msg = Store.newReturnAllITicketsMsg(tickets);
			EncDataSymm enc = EncDataSymm.encrypt(sessionKey, msg);
			sendMsg(Store.newReturnAllITickets(enc));
		} else {
			state = StateCard.IDLE;
			stop();
		}
	}

}
