package generated.eTicket.server.eTicketServer;

import swt.util.ByteArray;
import java.nio.charset.Charset;

/**
 * Class for T(L)V encoding of messages and other classes
 * 
 */
public class Coding {

	/**
	 * offset into encodeDestination for writing/reading
	 */
	private short encoding_length;

	/**
	 * array used for encoding/decoding
	 */
	private byte[] encodeDestination;

	/**
	 * Singleton instance
	 */
	private static Coding the_instance;

	private static Charset charset = Charset.forName("UTF-8");

	/**
	 * Singleton accessor
	 * 
	 * @return Singleton instance
	 */

	public static Coding getInstance() {
		if (the_instance == null)
			the_instance = new Coding();
		return the_instance;
	}
	/**
	 * return offset into encodeDestination
	 * 
	 * @return value of encoding_length
	 */
	public short getEncodingLength() {
		return encoding_length;
	}

	/**
	 * initializes encoding. This method must be called before encoding an object.
	 * The offset (encoding_length) is reset to 0 and encodeDestination is set to dest.
	 * 
	 * @param dest The (empty) destination array for encoding
	 */
	public void encodeInit(byte[] dest) {
		encodeDestination = dest;
		encoding_length = 0;
	}
	/**
	 * initializes decoding. The offset (encoding_length) is reset to 0.
	 */
	public void decodeInit() {
		encoding_length = 0;
	}

	/**
	 * initializes encoding and encodes an object.
	 * 
	 * @param o The object to encode
	 * @param dest destination array for the encoding
	 * @return length of the encoded object
	 */

	/**
	 * decodes an object beginning at given offset
	 * 
	 * @param in array containing encoded object
	 * @param offset where the encoded object begins
	 * @return the decoded object
	 */

	private byte decodeByte(byte[] in) throws ServiceException {
		if (in[encoding_length++] != Code.BYTE) {

			stop();
		}

		byte res = in[encoding_length++];
		return res;
	}

	private boolean decodeBoolean(byte[] in) throws ServiceException {
		if (in[encoding_length++] != Code.BOOLEAN) {
			stop();

		}
		boolean res = (in[encoding_length++] == 1 ? true : false);
		return res;
	}

	public Codeable decode(byte[] in, short offset) throws Exception {
		encoding_length = offset;
		return decode(in);
	}
	public byte[] stringToByte(String s) {
		byte[] b = s.getBytes(charset);
		byte[] ret = new byte[LengthConstants.LENGTHOFSTRING];
		int len = (b.length < LengthConstants.LENGTHOFSTRING)
				? b.length
				: LengthConstants.LENGTHOFSTRING;
		ByteArray.copy(b, 0, ret, 0, len);
		return ret;
	}

	public boolean isValidCardString(String s) {
		return (s.getBytes(charset).length <= LengthConstants.LENGTHOFSTRING);
	}

	public void encodeString(String i) {
		byte[] o = stringToByte(i);
		encodeDestination[encoding_length++] = Code.BYTEARRAY;
		ByteArray.setShort(encodeDestination, encoding_length,
				(short) (o.length));
		encoding_length += 2;
		ByteArray.copy(o, 0, encodeDestination, encoding_length, o.length);
		encoding_length += o.length;
	}

	private void compSizeString(String i) {
		byte[] o = stringToByte(i);
		encoding_length += 3;
		encoding_length += o.length;
	}

	public short encode(Codeable o, byte[] dest) throws Exception {
		encodeInit(dest);

		encode(o);
		return computeSize(o);

	}

	/**
	 * computes the encoding length of an object without actually encoding it.
	 * 
	 * @param o The object of which the encoding length should be computed
	 * @return encoding length
	 */

	public short computeSize(Codeable o) {
		encoding_length = 0;
		try {
			compSize(o);
			return encoding_length;
		} catch (Exception e) {
			return -1;
		}
	}

	public void encodeByte(byte o) {
		encodeDestination[encoding_length++] = Code.BYTE;
		encodeDestination[encoding_length++] = o;
	}

	private void encodeBoolean(boolean o) {
		encodeDestination[encoding_length++] = Code.BOOLEAN;
		encodeDestination[encoding_length++] = (byte) (o ? 1 : 0);
	}

	private void compSizeByte(byte o) {
		encoding_length += 2;
	}

	public void encodeInt(int i) {
		short o = (short) i;
		if (i != (int) o)
			System.err.println("int too big for cast into short.");
		encodeDestination[encoding_length++] = Code.SHORT;
		ByteArray.setShort(encodeDestination, encoding_length, o);
		encoding_length += 2;
	}

	private void compSizeInt(int i) {
		short o = (short) i;
		encoding_length += 3;
	}
	public int decodeInt(byte[] in) {
		short res = ByteArray.getShort(in, ++encoding_length);
		encoding_length += 2;
		return (int) res;
	}

	public byte[] decodeByteArray(byte[] in) {
		short len = ByteArray.getShort(in, ++encoding_length);
		byte[] dest = new byte[len];
		encoding_length += 2;
		ByteArray.copy(in, encoding_length, dest, 0, len);
		encoding_length += len;
		return dest;
	}

	public String byteToString(byte[] b) {
		int len = b.length;
		while (len > 0 && b[len - 1] == 0)
			len--;
		return new String(ByteArray.subarray(b, 0, len), charset);
	}

	public String decodeString(byte[] in) {
		short len = ByteArray.getShort(in, ++encoding_length);
		byte[] dest = new byte[len];
		encoding_length += 2;
		ByteArray.copy(in, encoding_length, dest, 0, len);
		encoding_length += len;
		String str = byteToString(dest);
		return str;
	}

	private void compSizeBoolean(boolean o) {
		encoding_length += 2;
	}

	public void encodeByteArray(byte[] o) {

		//byte[] o = stringToByte(i);

		encodeDestination[encoding_length++] = Code.BYTEARRAY;

		ByteArray.setShort(encodeDestination, encoding_length,
				(short) (o.length));
		encoding_length += 2;

		ByteArray.copy(o, 0, encodeDestination, encoding_length, o.length);

		encoding_length += o.length;
	}

	private void compSizeByteArray(byte[] o) {
		encoding_length += 3;
		encoding_length += o.length;
	}

	/**
	 * encodes an object without initializing destination and offset.
	 * 
	 * @param c object to encode
	 */

	public void encode(Codeable c) throws Exception {

		switch (c.getCode()) {
			case Code.REQTICKET :
				encodeReqTicket((ReqTicket) c);
				break;
			case Code.RESREQTICKET :
				encodeResReqTicket((ResReqTicket) c);
				break;
			case Code.BUYTICKETSUCCESSFUL :
				encodeBuyTicketSuccessful((BuyTicketSuccessful) c);
				break;
			case Code.CLIENTHELLO :
				encodeClientHello((ClientHello) c);
				break;
			case Code.SESSIONKEY :
				encodeSessionKey((SessionKey) c);
				break;
			case Code.SERVERHELLO :
				encodeServerHello((ServerHello) c);
				break;
			case Code.REQRECOVERY :
				encodeReqRecovery((ReqRecovery) c);
				break;
			case Code.RESREQRECOVERY :
				encodeResReqRecovery((ResReqRecovery) c);
				break;
			case Code.RECOVERYSUCCESSFUL :
				encodeRecoverySuccessful((RecoverySuccessful) c);
				break;
			case Code.HANDSHAKESUCCESSFUL :
				encodeHandshakeSuccessful((HandshakeSuccessful) c);
				break;
			case Code.TICKETRECEIVED :
				encodeTicketReceived((TicketReceived) c);
				break;
			case Code.TICKETRECEIVEDRECOVERY :
				encodeTicketReceivedRecovery((TicketReceivedRecovery) c);
				break;
			case Code.ENCDATASYMM :
				encodeEncDataSymm((EncDataSymm) c);
				break;
			case Code.BUYINFO :
				encodeBuyInfo((BuyInfo) c);
				break;
			case Code.RESREQTICKETDATA :
				encodeResReqTicketData((ResReqTicketData) c);
				break;
			case Code.PUBLICKEY :
				encodePublicKey((PublicKey) c);
				break;
			case Code.SIGNEDDATA :
				encodeSignedData((SignedData) c);
				break;
			case Code.ENCDATAASYMM :
				encodeEncDataAsymm((EncDataAsymm) c);
				break;
			case Code.CERTIFICATE :
				encodeCertificate((Certificate) c);
				break;
			case Code.CLIENTHELLOMSG :
				encodeClientHelloMsg((ClientHelloMsg) c);
				break;
			case Code.SESSIONKEYMSG :
				encodeSessionKeyMsg((SessionKeyMsg) c);
				break;
			case Code.SERVERHELLOMSG :
				encodeServerHelloMsg((ServerHelloMsg) c);
				break;
			case Code.RECOVERYINFO :
				encodeRecoveryInfo((RecoveryInfo) c);
				break;
			case Code.RECOVERYDATA :
				encodeRecoveryData((RecoveryData) c);
				break;
			case Code.TICKETRECEIVEDDATA :
				encodeTicketReceivedData((TicketReceivedData) c);
				break;
			case Code.TICKETRECEIVEDDATARECOVERY :
				encodeTicketReceivedDataRecovery((TicketReceivedDataRecovery) c);
				break;
			case Code.CCINFO :
				encodeCCInfo((CCInfo) c);
				break;
			case Code.TICKETINFO :
				encodeTicketInfo((TicketInfo) c);
				break;
			case Code.ID :
				encodeID((ID) c);
				break;
			case Code.NONCE :
				encodeNonce((Nonce) c);
				break;
			case Code.TICKET :
				encodeTicket((Ticket) c);
				break;
			case Code.SYMMKEY :
				encodeSymmKey((SymmKey) c);
				break;
			case Code.STAMP :
				encodeStamp((Stamp) c);
				break;
			case Code.DATETIME :
				encodeDateTime((DateTime) c);
				break;
			case Code.PRIVATEKEY :
				encodePrivateKey((PrivateKey) c);
				break;
			// never reached
			default :

				stop();

		}
	}

	public Codeable decode(byte[] in) throws Exception {
		switch (in[encoding_length]) {
			case Code.REQTICKET :
				return decodeReqTicket(in);
			case Code.RESREQTICKET :
				return decodeResReqTicket(in);
			case Code.BUYTICKETSUCCESSFUL :
				return decodeBuyTicketSuccessful(in);
			case Code.CLIENTHELLO :
				return decodeClientHello(in);
			case Code.SESSIONKEY :
				return decodeSessionKey(in);
			case Code.SERVERHELLO :
				return decodeServerHello(in);
			case Code.REQRECOVERY :
				return decodeReqRecovery(in);
			case Code.RESREQRECOVERY :
				return decodeResReqRecovery(in);
			case Code.RECOVERYSUCCESSFUL :
				return decodeRecoverySuccessful(in);
			case Code.HANDSHAKESUCCESSFUL :
				return decodeHandshakeSuccessful(in);
			case Code.TICKETRECEIVED :
				return decodeTicketReceived(in);
			case Code.TICKETRECEIVEDRECOVERY :
				return decodeTicketReceivedRecovery(in);
			case Code.ENCDATASYMM :
				return decodeEncDataSymm(in);
			case Code.BUYINFO :
				return decodeBuyInfo(in);
			case Code.RESREQTICKETDATA :
				return decodeResReqTicketData(in);
			case Code.PUBLICKEY :
				return decodePublicKey(in);
			case Code.SIGNEDDATA :
				return decodeSignedData(in);
			case Code.ENCDATAASYMM :
				return decodeEncDataAsymm(in);
			case Code.CERTIFICATE :
				return decodeCertificate(in);
			case Code.CLIENTHELLOMSG :
				return decodeClientHelloMsg(in);
			case Code.SESSIONKEYMSG :
				return decodeSessionKeyMsg(in);
			case Code.SERVERHELLOMSG :
				return decodeServerHelloMsg(in);
			case Code.RECOVERYINFO :
				return decodeRecoveryInfo(in);
			case Code.RECOVERYDATA :
				return decodeRecoveryData(in);
			case Code.TICKETRECEIVEDDATA :
				return decodeTicketReceivedData(in);
			case Code.TICKETRECEIVEDDATARECOVERY :
				return decodeTicketReceivedDataRecovery(in);
			case Code.CCINFO :
				return decodeCCInfo(in);
			case Code.TICKETINFO :
				return decodeTicketInfo(in);
			case Code.ID :
				return decodeID(in);
			case Code.NONCE :
				return decodeNonce(in);
			case Code.TICKET :
				return decodeTicket(in);
			case Code.SYMMKEY :
				return decodeSymmKey(in);
			case Code.STAMP :
				return decodeStamp(in);
			case Code.DATETIME :
				return decodeDateTime(in);
			case Code.PRIVATEKEY :
				return decodePrivateKey(in);
				// never reached
			default :
				stop();
		}
		return null;
	}

	public void compSize(Codeable c) throws Exception {
		switch (c.getCode()) {
			case Code.REQTICKET :

				compSizeReqTicket((ReqTicket) c);

				break;
			case Code.RESREQTICKET :

				compSizeResReqTicket((ResReqTicket) c);

				break;
			case Code.BUYTICKETSUCCESSFUL :

				compSizeBuyTicketSuccessful((BuyTicketSuccessful) c);

				break;
			case Code.CLIENTHELLO :

				compSizeClientHello((ClientHello) c);

				break;
			case Code.SESSIONKEY :

				compSizeSessionKey((SessionKey) c);

				break;
			case Code.SERVERHELLO :

				compSizeServerHello((ServerHello) c);

				break;
			case Code.REQRECOVERY :

				compSizeReqRecovery((ReqRecovery) c);

				break;
			case Code.RESREQRECOVERY :

				compSizeResReqRecovery((ResReqRecovery) c);

				break;
			case Code.RECOVERYSUCCESSFUL :

				compSizeRecoverySuccessful((RecoverySuccessful) c);

				break;
			case Code.HANDSHAKESUCCESSFUL :

				compSizeHandshakeSuccessful((HandshakeSuccessful) c);

				break;
			case Code.TICKETRECEIVED :

				compSizeTicketReceived((TicketReceived) c);

				break;
			case Code.TICKETRECEIVEDRECOVERY :

				compSizeTicketReceivedRecovery((TicketReceivedRecovery) c);

				break;
			case Code.ENCDATASYMM :

				compSizeEncDataSymm((EncDataSymm) c);

				break;
			case Code.BUYINFO :

				compSizeBuyInfo((BuyInfo) c);

				break;
			case Code.RESREQTICKETDATA :

				compSizeResReqTicketData((ResReqTicketData) c);

				break;
			case Code.PUBLICKEY :

				compSizePublicKey((PublicKey) c);

				break;
			case Code.SIGNEDDATA :

				compSizeSignedData((SignedData) c);

				break;
			case Code.ENCDATAASYMM :

				compSizeEncDataAsymm((EncDataAsymm) c);

				break;
			case Code.CERTIFICATE :

				compSizeCertificate((Certificate) c);

				break;
			case Code.CLIENTHELLOMSG :

				compSizeClientHelloMsg((ClientHelloMsg) c);

				break;
			case Code.SESSIONKEYMSG :

				compSizeSessionKeyMsg((SessionKeyMsg) c);

				break;
			case Code.SERVERHELLOMSG :

				compSizeServerHelloMsg((ServerHelloMsg) c);

				break;
			case Code.RECOVERYINFO :

				compSizeRecoveryInfo((RecoveryInfo) c);

				break;
			case Code.RECOVERYDATA :

				compSizeRecoveryData((RecoveryData) c);

				break;
			case Code.TICKETRECEIVEDDATA :

				compSizeTicketReceivedData((TicketReceivedData) c);

				break;
			case Code.TICKETRECEIVEDDATARECOVERY :

				compSizeTicketReceivedDataRecovery((TicketReceivedDataRecovery) c);

				break;
			case Code.CCINFO :

				compSizeCCInfo((CCInfo) c);

				break;
			case Code.TICKETINFO :

				compSizeTicketInfo((TicketInfo) c);

				break;
			case Code.ID :

				compSizeID((ID) c);

				break;
			case Code.NONCE :

				compSizeNonce((Nonce) c);

				break;
			case Code.TICKET :

				compSizeTicket((Ticket) c);

				break;
			case Code.SYMMKEY :

				compSizeSymmKey((SymmKey) c);

				break;
			case Code.STAMP :

				compSizeStamp((Stamp) c);

				break;
			case Code.DATETIME :

				compSizeDateTime((DateTime) c);

				break;
			case Code.PRIVATEKEY :

				compSizePrivateKey((PrivateKey) c);

				break;
			// never reached
			default :

				stop();

		}
	}

	public PlainData decodePlainData(byte[] in, byte expectedType)
			throws Exception {
		encoding_length = 0;
		PlainData p = null;
		try {
			if (in[encoding_length] != expectedType) {
				stop();

			}
			switch (in[encoding_length]) {

				case Code.BUYINFO :
					p = decodeBuyInfo(in);
					break;
				case Code.RESREQTICKETDATA :
					p = decodeResReqTicketData(in);
					break;
				case Code.CLIENTHELLOMSG :
					p = decodeClientHelloMsg(in);
					break;
				case Code.SESSIONKEYMSG :
					p = decodeSessionKeyMsg(in);
					break;
				case Code.SERVERHELLOMSG :
					p = decodeServerHelloMsg(in);
					break;
				case Code.RECOVERYINFO :
					p = decodeRecoveryInfo(in);
					break;
				case Code.RECOVERYDATA :
					p = decodeRecoveryData(in);
					break;
				case Code.TICKETRECEIVEDDATA :
					p = decodeTicketReceivedData(in);
					break;
				case Code.TICKETRECEIVEDDATARECOVERY :
					p = decodeTicketReceivedDataRecovery(in);
					break;
				case Code.TICKET :
					p = decodeTicket(in);
					break;

			}
		} catch (ArrayIndexOutOfBoundsException e) {
			stop();

		}

		if (p == null) {
			stop();

		}
		return p;
	}

	public void encodeReqTicket(ReqTicket c) throws Exception {
		encodeDestination[encoding_length++] = Code.REQTICKET;

		encode(c.getMsg());
	}

	public ReqTicket decodeReqTicket(byte[] in) throws Exception {
		ReqTicket c = new ReqTicket();
		encoding_length++;
		c.setMsg(((EncDataSymm) decode(in)));
		return c;
	}

	private void compSizeReqTicket(ReqTicket c) throws Exception {
		encoding_length++;
		compSize(c.getMsg());
	}

	public void encodeResReqTicket(ResReqTicket c) throws Exception {
		encodeDestination[encoding_length++] = Code.RESREQTICKET;

		encode(c.getMsg());
	}

	public ResReqTicket decodeResReqTicket(byte[] in) throws Exception {
		ResReqTicket c = new ResReqTicket();
		encoding_length++;
		c.setMsg(((EncDataSymm) decode(in)));
		return c;
	}

	private void compSizeResReqTicket(ResReqTicket c) throws Exception {
		encoding_length++;
		compSize(c.getMsg());
	}

	public void encodeBuyTicketSuccessful(BuyTicketSuccessful c)
			throws Exception {
		encodeDestination[encoding_length++] = Code.BUYTICKETSUCCESSFUL;
	}

	public BuyTicketSuccessful decodeBuyTicketSuccessful(byte[] in)
			throws Exception {
		BuyTicketSuccessful c = new BuyTicketSuccessful();
		encoding_length++;
		return c;
	}

	private void compSizeBuyTicketSuccessful(BuyTicketSuccessful c)
			throws Exception {
		encoding_length++;
	}

	public void encodeClientHello(ClientHello c) throws Exception {
		encodeDestination[encoding_length++] = Code.CLIENTHELLO;

		encode(c.getMsg());

		encode(c.getCertCard());

		encode(c.getPublicKeyCard());
	}

	public ClientHello decodeClientHello(byte[] in) throws Exception {
		ClientHello c = new ClientHello();
		encoding_length++;
		c.setMsg(((EncDataAsymm) decode(in)));
		c.setCertCard(((SignedData) decode(in)));
		c.setPublicKeyCard(((PublicKey) decode(in)));
		return c;
	}

	private void compSizeClientHello(ClientHello c) throws Exception {
		encoding_length++;
		compSize(c.getMsg());
		compSize(c.getCertCard());
		compSize(c.getPublicKeyCard());
	}

	public void encodeSessionKey(SessionKey c) throws Exception {
		encodeDestination[encoding_length++] = Code.SESSIONKEY;

		encode(c.getMsg());
	}

	public SessionKey decodeSessionKey(byte[] in) throws Exception {
		SessionKey c = new SessionKey();
		encoding_length++;
		c.setMsg(((EncDataAsymm) decode(in)));
		return c;
	}

	private void compSizeSessionKey(SessionKey c) throws Exception {
		encoding_length++;
		compSize(c.getMsg());
	}

	public void encodeServerHello(ServerHello c) throws Exception {
		encodeDestination[encoding_length++] = Code.SERVERHELLO;

		encode(c.getMsg());
	}

	public ServerHello decodeServerHello(byte[] in) throws Exception {
		ServerHello c = new ServerHello();
		encoding_length++;
		c.setMsg(((EncDataAsymm) decode(in)));
		return c;
	}

	private void compSizeServerHello(ServerHello c) throws Exception {
		encoding_length++;
		compSize(c.getMsg());
	}

	public void encodeReqRecovery(ReqRecovery c) throws Exception {
		encodeDestination[encoding_length++] = Code.REQRECOVERY;

		encode(c.getRecoveryInfo());
	}

	public ReqRecovery decodeReqRecovery(byte[] in) throws Exception {
		ReqRecovery c = new ReqRecovery();
		encoding_length++;
		c.setRecoveryInfo(((EncDataSymm) decode(in)));
		return c;
	}

	private void compSizeReqRecovery(ReqRecovery c) throws Exception {
		encoding_length++;
		compSize(c.getRecoveryInfo());
	}

	public void encodeResReqRecovery(ResReqRecovery c) throws Exception {
		encodeDestination[encoding_length++] = Code.RESREQRECOVERY;

		encode(c.getRecoveryData());
	}

	public ResReqRecovery decodeResReqRecovery(byte[] in) throws Exception {
		ResReqRecovery c = new ResReqRecovery();
		encoding_length++;
		c.setRecoveryData(((EncDataSymm) decode(in)));
		return c;
	}

	private void compSizeResReqRecovery(ResReqRecovery c) throws Exception {
		encoding_length++;
		compSize(c.getRecoveryData());
	}

	public void encodeRecoverySuccessful(RecoverySuccessful c) throws Exception {
		encodeDestination[encoding_length++] = Code.RECOVERYSUCCESSFUL;
	}

	public RecoverySuccessful decodeRecoverySuccessful(byte[] in)
			throws Exception {
		RecoverySuccessful c = new RecoverySuccessful();
		encoding_length++;
		return c;
	}

	private void compSizeRecoverySuccessful(RecoverySuccessful c)
			throws Exception {
		encoding_length++;
	}

	public void encodeHandshakeSuccessful(HandshakeSuccessful c)
			throws Exception {
		encodeDestination[encoding_length++] = Code.HANDSHAKESUCCESSFUL;
	}

	public HandshakeSuccessful decodeHandshakeSuccessful(byte[] in)
			throws Exception {
		HandshakeSuccessful c = new HandshakeSuccessful();
		encoding_length++;
		return c;
	}

	private void compSizeHandshakeSuccessful(HandshakeSuccessful c)
			throws Exception {
		encoding_length++;
	}

	public void encodeTicketReceived(TicketReceived c) throws Exception {
		encodeDestination[encoding_length++] = Code.TICKETRECEIVED;

		encode(c.getMsg());
	}

	public TicketReceived decodeTicketReceived(byte[] in) throws Exception {
		TicketReceived c = new TicketReceived();
		encoding_length++;
		c.setMsg(((EncDataSymm) decode(in)));
		return c;
	}

	private void compSizeTicketReceived(TicketReceived c) throws Exception {
		encoding_length++;
		compSize(c.getMsg());
	}

	public void encodeTicketReceivedRecovery(TicketReceivedRecovery c)
			throws Exception {
		encodeDestination[encoding_length++] = Code.TICKETRECEIVEDRECOVERY;

		encode(c.getMsg());
	}

	public TicketReceivedRecovery decodeTicketReceivedRecovery(byte[] in)
			throws Exception {
		TicketReceivedRecovery c = new TicketReceivedRecovery();
		encoding_length++;
		c.setMsg(((EncDataSymm) decode(in)));
		return c;
	}

	private void compSizeTicketReceivedRecovery(TicketReceivedRecovery c)
			throws Exception {
		encoding_length++;
		compSize(c.getMsg());
	}

	private void encodeEncDataSymm(EncDataSymm c) {

		encodeDestination[encoding_length++] = Code.ENCDATASYMM;
		encodeByteArray(c.encrypted);

		encodeInt(c.plainlength);
		encodeInt(c.enclength);

		encodeByte(c.plainDataType);
	}

	public EncDataSymm decodeEncDataSymm(byte[] in) throws ServiceException {
		if (in[encoding_length++] != Code.ENCDATASYMM)
			stop();
		EncDataSymm into = new EncDataSymm();
		//public byte[] encrypted;
		//public short plainlength;
		//public short enclength;
		//public byte plainDataType;
		into.encrypted = decodeByteArray(in);
		into.plainlength = decodeInt(in);
		into.enclength = decodeInt(in);
		into.plainDataType = decodeByte(in);
		return into;
	}

	private void compSizeEncDataSymm(EncDataSymm c) {
		encoding_length++;
		compSizeByteArray(c.encrypted);
		encoding_length += 8; // short + short + byte
	}

	public void encodeBuyInfo(BuyInfo c) throws Exception {
		encodeDestination[encoding_length++] = Code.BUYINFO;

		encode(c.getCcInfo());

		encode(c.getTicketInfo());

		encode(c.getCardID());

		encode(c.getTransactionId());
	}

	public BuyInfo decodeBuyInfo(byte[] in) throws Exception {
		BuyInfo c = new BuyInfo();
		encoding_length++;
		c.setCcInfo(((CCInfo) decode(in)));
		c.setTicketInfo(((TicketInfo) decode(in)));
		c.setCardID(((ID) decode(in)));
		c.setTransactionId(((Nonce) decode(in)));
		return c;
	}

	private void compSizeBuyInfo(BuyInfo c) throws Exception {
		encoding_length++;
		compSize(c.getCcInfo());
		compSize(c.getTicketInfo());
		compSize(c.getCardID());
		compSize(c.getTransactionId());
	}

	public void encodeResReqTicketData(ResReqTicketData c) throws Exception {
		encodeDestination[encoding_length++] = Code.RESREQTICKETDATA;

		encode(c.getTicket());

		encode(c.getTransactionId());

		encode(c.getNonceDeleteTicket());
	}

	public ResReqTicketData decodeResReqTicketData(byte[] in) throws Exception {
		ResReqTicketData c = new ResReqTicketData();
		encoding_length++;
		c.setTicket(((Ticket) decode(in)));
		c.setTransactionId(((Nonce) decode(in)));
		c.setNonceDeleteTicket(((Nonce) decode(in)));
		return c;
	}

	private void compSizeResReqTicketData(ResReqTicketData c) throws Exception {
		encoding_length++;
		compSize(c.getTicket());
		compSize(c.getTransactionId());
		compSize(c.getNonceDeleteTicket());
	}

	private void encodePublicKey(PublicKey c) throws Exception {
		encodeDestination[encoding_length++] = Code.PUBLICKEY;
		encodeByteArray(c.getModulus());

	}

	public PublicKey decodePublicKey(byte[] in) throws Exception {
		if (in[encoding_length++] != Code.PUBLICKEY)
			stop();
		PublicKey res = new PublicKey();
		res.setKey(decodeByteArray(in));
		return res;
	}

	private void compSizePublicKey(PublicKey c) throws Exception {
		encoding_length++;
		compSizeByteArray(c.getModulus());

	}

	public void encodeSignedData(SignedData c) throws Exception {
		encodeDestination[encoding_length++] = Code.SIGNEDDATA;

		encodeByteArray((byte[]) c.getSigned());
	}

	public SignedData decodeSignedData(byte[] in) throws Exception {
		SignedData c = new SignedData();
		encoding_length++;
		c.setSigned(decodeByteArray(in));
		return c;
	}

	private void compSizeSignedData(SignedData c) throws Exception {
		encoding_length++;

		compSizeByteArray(c.getSigned());
	}

	private void encodeEncDataAsymm(EncDataAsymm c) {

		encodeDestination[encoding_length++] = Code.ENCDATAASYMM;
		encodeByteArray(c.encrypted);

		encodeInt(c.plainlength);
		encodeInt(c.enclength);

		encodeByte(c.plainDataType);
	}

	public EncDataAsymm decodeEncDataAsymm(byte[] in) throws Exception {
		if (in[encoding_length++] != Code.ENCDATAASYMM)
			stop();
		EncDataAsymm into = new EncDataAsymm();
		//public byte[] encrypted;
		//public short plainlength;
		//public short enclength;
		//public byte plainDataType;
		into.encrypted = decodeByteArray(in);
		into.plainlength = decodeInt(in);
		into.enclength = decodeInt(in);
		into.plainDataType = decodeByte(in);
		return into;
	}

	private void compSizeEncDataAsymm(EncDataAsymm c) {
		encoding_length++;
		compSizeByteArray(c.encrypted);
		encoding_length += 8; // short + short + byte
	}

	public void encodeCertificate(Certificate c) throws Exception {
		encodeDestination[encoding_length++] = Code.CERTIFICATE;

		encode(c.getPublicKey());
	}

	public Certificate decodeCertificate(byte[] in) throws Exception {
		Certificate c = new Certificate();
		encoding_length++;
		c.setPublicKey(((PublicKey) decode(in)));
		return c;
	}

	private void compSizeCertificate(Certificate c) throws Exception {
		encoding_length++;
		compSize(c.getPublicKey());
	}

	public void encodeClientHelloMsg(ClientHelloMsg c) throws Exception {
		encodeDestination[encoding_length++] = Code.CLIENTHELLOMSG;

		encode(c.getNonceCard());
	}

	public ClientHelloMsg decodeClientHelloMsg(byte[] in) throws Exception {
		ClientHelloMsg c = new ClientHelloMsg();
		encoding_length++;
		c.setNonceCard(((Nonce) decode(in)));
		return c;
	}

	private void compSizeClientHelloMsg(ClientHelloMsg c) throws Exception {
		encoding_length++;
		compSize(c.getNonceCard());
	}

	public void encodeSessionKeyMsg(SessionKeyMsg c) throws Exception {
		encodeDestination[encoding_length++] = Code.SESSIONKEYMSG;

		encode(c.getNonceService());

		encode(c.getSessionKey());
	}

	public SessionKeyMsg decodeSessionKeyMsg(byte[] in) throws Exception {
		SessionKeyMsg c = new SessionKeyMsg();
		encoding_length++;
		c.setNonceService(((Nonce) decode(in)));
		c.setSessionKey(((SymmKey) decode(in)));
		return c;
	}

	private void compSizeSessionKeyMsg(SessionKeyMsg c) throws Exception {
		encoding_length++;
		compSize(c.getNonceService());
		compSize(c.getSessionKey());
	}

	public void encodeServerHelloMsg(ServerHelloMsg c) throws Exception {
		encodeDestination[encoding_length++] = Code.SERVERHELLOMSG;

		encode(c.getNonceService());

		encode(c.getNonceCard());
	}

	public ServerHelloMsg decodeServerHelloMsg(byte[] in) throws Exception {
		ServerHelloMsg c = new ServerHelloMsg();
		encoding_length++;
		c.setNonceService(((Nonce) decode(in)));
		c.setNonceCard(((Nonce) decode(in)));
		return c;
	}

	private void compSizeServerHelloMsg(ServerHelloMsg c) throws Exception {
		encoding_length++;
		compSize(c.getNonceService());
		compSize(c.getNonceCard());
	}

	public void encodeRecoveryInfo(RecoveryInfo c) throws Exception {
		encodeDestination[encoding_length++] = Code.RECOVERYINFO;

		encode(c.getCardId());

		encode(c.getTransactionId());
	}

	public RecoveryInfo decodeRecoveryInfo(byte[] in) throws Exception {
		RecoveryInfo c = new RecoveryInfo();
		encoding_length++;
		c.setCardId(((ID) decode(in)));
		c.setTransactionId(((Nonce) decode(in)));
		return c;
	}

	private void compSizeRecoveryInfo(RecoveryInfo c) throws Exception {
		encoding_length++;
		compSize(c.getCardId());
		compSize(c.getTransactionId());
	}

	public void encodeRecoveryData(RecoveryData c) throws Exception {
		encodeDestination[encoding_length++] = Code.RECOVERYDATA;

		encode(c.getTicket());

		encode(c.getTransactionId());

		encode(c.getNonceDeleteTicket());
	}

	public RecoveryData decodeRecoveryData(byte[] in) throws Exception {
		RecoveryData c = new RecoveryData();
		encoding_length++;
		c.setTicket(((Ticket) decode(in)));
		c.setTransactionId(((Nonce) decode(in)));
		c.setNonceDeleteTicket(((Nonce) decode(in)));
		return c;
	}

	private void compSizeRecoveryData(RecoveryData c) throws Exception {
		encoding_length++;
		compSize(c.getTicket());
		compSize(c.getTransactionId());
		compSize(c.getNonceDeleteTicket());
	}

	public void encodeTicketReceivedData(TicketReceivedData c) throws Exception {
		encodeDestination[encoding_length++] = Code.TICKETRECEIVEDDATA;

		encode(c.getCardId());

		encode(c.getNonceDeleteTicket());
	}

	public TicketReceivedData decodeTicketReceivedData(byte[] in)
			throws Exception {
		TicketReceivedData c = new TicketReceivedData();
		encoding_length++;
		c.setCardId(((ID) decode(in)));
		c.setNonceDeleteTicket(((Nonce) decode(in)));
		return c;
	}

	private void compSizeTicketReceivedData(TicketReceivedData c)
			throws Exception {
		encoding_length++;
		compSize(c.getCardId());
		compSize(c.getNonceDeleteTicket());
	}

	public void encodeTicketReceivedDataRecovery(TicketReceivedDataRecovery c)
			throws Exception {
		encodeDestination[encoding_length++] = Code.TICKETRECEIVEDDATARECOVERY;

		encode(c.getCardId());

		encode(c.getNonceDeleteTicket());
	}

	public TicketReceivedDataRecovery decodeTicketReceivedDataRecovery(byte[] in)
			throws Exception {
		TicketReceivedDataRecovery c = new TicketReceivedDataRecovery();
		encoding_length++;
		c.setCardId(((ID) decode(in)));
		c.setNonceDeleteTicket(((Nonce) decode(in)));
		return c;
	}

	private void compSizeTicketReceivedDataRecovery(TicketReceivedDataRecovery c)
			throws Exception {
		encoding_length++;
		compSize(c.getCardId());
		compSize(c.getNonceDeleteTicket());
	}

	public void encodeCCInfo(CCInfo c) throws Exception {
		encodeDestination[encoding_length++] = Code.CCINFO;

		encodeString((String) c.getCCHolderName());

		encodeString((String) c.getCCSecNumber());

		encodeString((String) c.getCCNumber());

		encodeString((String) c.getCCType());

		encodeString((String) c.getCCExpDate());
	}

	public CCInfo decodeCCInfo(byte[] in) throws Exception {
		CCInfo c = new CCInfo();
		encoding_length++;
		c.setCCHolderName(decodeString(in));
		c.setCCSecNumber(decodeString(in));
		c.setCCNumber(decodeString(in));
		c.setCCType(decodeString(in));
		c.setCCExpDate(decodeString(in));
		return c;
	}

	private void compSizeCCInfo(CCInfo c) throws Exception {
		encoding_length++;

		compSizeString(c.getCCHolderName());

		compSizeString(c.getCCSecNumber());

		compSizeString(c.getCCNumber());

		compSizeString(c.getCCType());

		compSizeString(c.getCCExpDate());
	}

	public void encodeTicketInfo(TicketInfo c) throws Exception {
		encodeDestination[encoding_length++] = Code.TICKETINFO;

		encodeString((String) c.getDeparture());

		encodeString((String) c.getDestination());

		encodeInt((int) c.getTrainClass());

		encodeString((String) c.getExpiration());
	}

	public TicketInfo decodeTicketInfo(byte[] in) throws Exception {
		TicketInfo c = new TicketInfo();
		encoding_length++;
		c.setDeparture(decodeString(in));
		c.setDestination(decodeString(in));
		c.setTrainClass(decodeInt(in));
		c.setExpiration(decodeString(in));
		return c;
	}

	private void compSizeTicketInfo(TicketInfo c) throws Exception {
		encoding_length++;

		compSizeString(c.getDeparture());

		compSizeString(c.getDestination());

		compSizeInt(c.getTrainClass());

		compSizeString(c.getExpiration());
	}

	public void encodeID(ID c) throws Exception {
		encodeDestination[encoding_length++] = Code.ID;

		encodeInt((int) c.getId());
	}

	public ID decodeID(byte[] in) throws Exception {
		ID c = new ID();
		encoding_length++;
		c.setId(decodeInt(in));
		return c;
	}

	private void compSizeID(ID c) throws Exception {
		encoding_length++;

		compSizeInt(c.getId());
	}

	public void encodeNonce(Nonce c) throws Exception {
		encodeDestination[encoding_length++] = Code.NONCE;

		encodeByteArray((byte[]) c.getNonce());
	}

	public Nonce decodeNonce(byte[] in) throws Exception {
		Nonce c = new Nonce();
		encoding_length++;
		c.setNonce(decodeByteArray(in));
		return c;
	}

	private void compSizeNonce(Nonce c) throws Exception {
		encoding_length++;

		compSizeByteArray(c.getNonce());
	}

	public void encodeTicket(Ticket c) throws Exception {
		encodeDestination[encoding_length++] = Code.TICKET;

		encode(c.getId());

		encode(c.getTicketInfo());

		encode(c.getStamp());
	}

	public Ticket decodeTicket(byte[] in) throws Exception {
		Ticket c = new Ticket();
		encoding_length++;
		c.setId(((ID) decode(in)));
		c.setTicketInfo(((TicketInfo) decode(in)));
		c.setStamp(((Stamp) decode(in)));
		return c;
	}

	private void compSizeTicket(Ticket c) throws Exception {
		encoding_length++;
		compSize(c.getId());
		compSize(c.getTicketInfo());
		compSize(c.getStamp());
	}

	public void encodeSymmKey(SymmKey c) throws Exception {
		encodeDestination[encoding_length++] = Code.SYMMKEY;

		encodeByteArray((byte[]) c.getKey());
	}

	public SymmKey decodeSymmKey(byte[] in) throws Exception {
		SymmKey c = new SymmKey();
		encoding_length++;
		c.setKey(decodeByteArray(in));
		return c;
	}

	private void compSizeSymmKey(SymmKey c) throws Exception {
		encoding_length++;

		compSizeByteArray(c.getKey());
	}

	public void encodeStamp(Stamp c) throws Exception {
		encodeDestination[encoding_length++] = Code.STAMP;

		encodeBoolean((boolean) c.getStamped());

		encode(c.getStampTime());
	}

	public Stamp decodeStamp(byte[] in) throws Exception {
		Stamp c = new Stamp();
		encoding_length++;
		c.setStamped(decodeBoolean(in));
		c.setStampTime(((DateTime) decode(in)));
		return c;
	}

	private void compSizeStamp(Stamp c) throws Exception {
		encoding_length++;

		compSizeBoolean(c.getStamped());
		compSize(c.getStampTime());
	}

	public void encodeDateTime(DateTime c) throws Exception {
		encodeDestination[encoding_length++] = Code.DATETIME;

		encodeString((String) c.getDateTime());
	}

	public DateTime decodeDateTime(byte[] in) throws Exception {
		DateTime c = new DateTime();
		encoding_length++;
		c.setDateTime(decodeString(in));
		return c;
	}

	private void compSizeDateTime(DateTime c) throws Exception {
		encoding_length++;

		compSizeString(c.getDateTime());
	}

	public void encodePrivateKey(PrivateKey c) throws Exception {
		encodeDestination[encoding_length++] = Code.PRIVATEKEY;

		encodeByteArray(c.getModulus());
		encodeByteArray(c.getPrivexponent());

	}

	public PrivateKey decodePrivateKey(byte[] in) throws Exception {
		if (in[encoding_length++] != Code.PRIVATEKEY)
			stop();
		PrivateKey res = new PrivateKey();
		res.setKey(decodeByteArray(in), decodeByteArray(in));
		return res;
	}

	private void compSizePrivateKey(PrivateKey c) throws Exception {
		encoding_length++;

		compSizeByteArray(c.getModulus());
		compSizeByteArray(c.getPrivexponent());

	}

	private void stop() throws ServiceException {
		throw new ServiceException("Coding");
	}
}
