package generated.eTicket.server.eTicketServer;

import java.io.FileInputStream;
import java.net.InetSocketAddress;
import java.security.KeyStore;
import javax.jws.WebMethod;
import javax.jws.WebService;
import javax.net.ssl.*;
import javax.swing.JOptionPane;
import javax.xml.ws.Endpoint;
import javax.xml.ws.soap.Addressing;
import javax.xml.ws.BindingProvider;

import com.sun.net.httpserver.*;
import com.sun.xml.ws.developer.Stateful;
import com.sun.xml.ws.developer.StatefulWebServiceManager;

/**
 *
 * @author Marian Borek
 */

@Stateful
@Addressing
@WebService
public class ETicketServer {

	private PrivateKey signKey;

	private PublicKey certVerifyKey;

	private PrivateKey privateKeyServer;

	private SymmKey sessionKey;

	private Nonce nonceServer;

	private Nonce nonceDeleteTicket;

	private ID ticketID;

	private ListOfID issuedTicketIDs;

	private byte state;

	private ListOfCardTicket tickets;

	public static StatefulWebServiceManager<ETicketServer> manager;

	private Endpoint endpoint;
	private HttpsServer httpsServer;

	public ETicketServer() throws java.lang.Exception {

		signKey = new PrivateKey();
		certVerifyKey = new PublicKey();
		privateKeyServer = new PrivateKey();
		sessionKey = new SymmKey();
		nonceServer = new Nonce();
		nonceDeleteTicket = new Nonce();
		ticketID = new ID();
		issuedTicketIDs = new ListOfID();
		state = StateService.IDLE;
		tickets = new ListOfCardTicket();

	}

	public ETicketServer(PublicKey certVerifyKey, PrivateKey privateKeyServer,
			PrivateKey signKey, byte state) throws java.lang.Exception {

		this.certVerifyKey = certVerifyKey;
		this.privateKeyServer = privateKeyServer;
		this.signKey = signKey;
		this.state = state;

		sessionKey = new SymmKey();
		nonceServer = new Nonce();
		nonceDeleteTicket = new Nonce();
		ticketID = new ID();
		issuedTicketIDs = new ListOfID();
		tickets = new ListOfCardTicket();

	}

	@WebMethod
	public synchronized void closeService() {
		manager.unexport(this);
	}

	private Message sendMsg(Message msg) {

		return msg;

	}

	private Message sendMsg(Message msg, int port) throws java.lang.Exception {

		return sendMsg(msg, port, false, false);

	}
	private Message sendMsg(Message msg, int port,
			Boolean openSessionBeforeSend, Boolean closeSessionAfterReceive)
			throws java.lang.Exception {

		Message response;

		switch (port) {
			default :
				stop();
				return null;
		}
	}

	private Message processMessage(Message inmsg) throws java.lang.Exception {
		switch (inmsg.getCode()) {
			case Code.REQTICKET :
				return processReqTicket((ReqTicket) inmsg);
			case Code.TICKETRECEIVED :
				return processTicketReceived((TicketReceived) inmsg);
			case Code.CLIENTHELLO :
				return processClientHello((ClientHello) inmsg);
			case Code.SESSIONKEY :
				return processSessionKey((SessionKey) inmsg);
			case Code.REQRECOVERY :
				return processReqRecovery((ReqRecovery) inmsg);
			case Code.TICKETRECEIVEDRECOVERY :
				return processTicketReceivedRecovery((TicketReceivedRecovery) inmsg);
			default :
				stop();
				return null;
		}
	}

	@WebMethod
	public synchronized MessageWrapper process(MessageWrapper msg)
			throws ServiceException {
		MessageWrapper m = null;
		try {
			m = new MessageWrapper(processMessage(msg.getMsg()));
		} catch (java.lang.Exception e) {
			e.printStackTrace();
			if (e instanceof ServiceException)
				throw (ServiceException) e;
			else
				throw new ServiceException("Error when calling process method");
		}
		return m;
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	private Message processReqTicket(ReqTicket inmsg)
			throws java.lang.Exception {
		synchronized (manager) {
			EncDataSymm enc = inmsg.getMsg();
			if (state == StateService.getAUTHENTICATED()) {
				BuyInfo msg = (BuyInfo) (EncDataSymm.decrypt(sessionKey, enc));
				tickets.remove(msg.getCardID());
				boolean paymentSuccessful = pay(msg.getCcInfo());
				if (paymentSuccessful) {
					ID ticketID = generateID();
					issuedTicketIDs.add(ticketID);
					Ticket ticket = createTicket(ticketID, msg.getTicketInfo());
					CardTicket cardTicket = new CardTicket(msg.getCardID(),
							ticket);
					tickets.set(cardTicket);
					setNonceDeleteTicket(Nonce.generateNonce());
					ResReqTicketData msgResponse = new ResReqTicketData(ticket,
							msg.getTransactionId(), nonceDeleteTicket);
					EncDataSymm encResponse = EncDataSymm.encrypt(sessionKey,
							msgResponse);
					setState(StateService.getEXP_RECEIVED());
					return sendMsg(new ResReqTicket(encResponse));
				} else {
					setState(StateService.getIDLE());
					stop();

					return null;
				}
			} else {
				setState(StateService.getIDLE());
				stop();

				return null;
			}
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	private Message processTicketReceived(TicketReceived inmsg)
			throws java.lang.Exception {
		synchronized (manager) {
			EncDataSymm enc = inmsg.getMsg();
			if (state == StateService.getEXP_RECEIVED()) {
				TicketReceivedData trd = (TicketReceivedData) (EncDataSymm
						.decrypt(sessionKey, enc));
				if (trd.getNonceDeleteTicket().equals(
						this.getNonceDeleteTicket())) {
					tickets.remove(trd.getCardId());
					setState(StateService.getIDLE());
					return sendMsg(new BuyTicketSuccessful());
				} else {
					setState(StateService.getIDLE());
					stop();

					return null;
				}
			} else {
				setState(StateService.getIDLE());
				stop();

				return null;
			}
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	private Message processClientHello(ClientHello inmsg)
			throws java.lang.Exception {
		synchronized (manager) {
			EncDataAsymm enc = inmsg.getMsg();
			SignedData certCard = inmsg.getCertCard();
			PublicKey publicKeyCard = inmsg.getPublicKeyCard();
			if (state == StateService.getIDLE()) {
				Certificate certCardData = new Certificate(publicKeyCard);
				boolean correctCert = SignedData.verify(certVerifyKey,
						certCard, certCardData);
				if (correctCert) {
					ClientHelloMsg chm = (ClientHelloMsg) (EncDataAsymm
							.decrypt(privateKeyServer, enc));
					setNonceServer(Nonce.generateNonce());
					ServerHelloMsg shm = new ServerHelloMsg(nonceServer,
							chm.getNonceCard());
					EncDataAsymm enc1 = EncDataAsymm
							.encrypt(publicKeyCard, shm);
					setState(StateService.getWAIT_FOR_SESSION_KEY());
					return sendMsg(new ServerHello(enc1));
				} else {
					setState(StateService.getIDLE());
					stop();

					return null;
				}
			} else {
				setState(StateService.getIDLE());
				stop();

				return null;
			}
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	private Message processSessionKey(SessionKey inmsg)
			throws java.lang.Exception {
		synchronized (manager) {
			EncDataAsymm enc = inmsg.getMsg();
			if (state == StateService.getWAIT_FOR_SESSION_KEY()) {
				SessionKeyMsg skm = (SessionKeyMsg) (EncDataAsymm.decrypt(
						privateKeyServer, enc));
				if (skm.getNonceService().equals(nonceServer)) {
					setSessionKey(skm.getSessionKey());
					setState(StateService.getAUTHENTICATED());
					return sendMsg(new HandshakeSuccessful());
				} else {
					setState(StateService.getIDLE());
					stop();

					return null;
				}
			} else {
				setState(StateService.getIDLE());
				stop();

				return null;
			}
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	private Message processReqRecovery(ReqRecovery inmsg)
			throws java.lang.Exception {
		synchronized (manager) {
			EncDataSymm enc = inmsg.getRecoveryInfo();
			if (state == StateService.getAUTHENTICATED()) {
				RecoveryInfo msg = (RecoveryInfo) (EncDataSymm.decrypt(
						sessionKey, enc));
				CardTicket cardTicket = tickets.get(msg.getCardId());
				setNonceDeleteTicket(Nonce.generateNonce());
				RecoveryData recoveryData = new RecoveryData(
						cardTicket.getTicket(), msg.getTransactionId(),
						nonceDeleteTicket);
				EncDataSymm resEnc = EncDataSymm.encrypt(sessionKey,
						recoveryData);
				setState(StateService.getEXP_RECEIVED());
				return sendMsg(new ResReqRecovery(resEnc));
			} else {
				setState(StateService.getIDLE());
				stop();

				return null;
			}
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	private Message processTicketReceivedRecovery(TicketReceivedRecovery inmsg)
			throws java.lang.Exception {
		synchronized (manager) {
			EncDataSymm enc = inmsg.getMsg();
			if (state == StateService.getEXP_RECEIVED()) {
				TicketReceivedDataRecovery trd = (TicketReceivedDataRecovery) (EncDataSymm
						.decrypt(sessionKey, enc));
				if (trd.getNonceDeleteTicket().equals(
						this.getNonceDeleteTicket())) {
					tickets.remove(trd.getCardId());
					setState(StateService.getIDLE());
					return sendMsg(new RecoverySuccessful());
				} else {
					setState(StateService.getIDLE());
					stop();

					return null;
				}
			} else {
				setState(StateService.getIDLE());
				stop();

				return null;
			}
		}
	}

	public PrivateKey getSignKey() {
		return signKey;
	}

	public PublicKey getCertVerifyKey() {
		return certVerifyKey;
	}

	public PrivateKey getPrivateKeyServer() {
		return privateKeyServer;
	}

	public SymmKey getSessionKey() {
		return sessionKey;
	}

	public Nonce getNonceServer() {
		return nonceServer;
	}

	public Nonce getNonceDeleteTicket() {
		return nonceDeleteTicket;
	}

	public ID getTicketID() {
		return ticketID;
	}

	public ListOfID getIssuedTicketIDs() {
		return issuedTicketIDs;
	}

	public byte getState() {
		return state;
	}

	public ListOfCardTicket getTickets() {
		return tickets;
	}

	public void setSignKey(PrivateKey signKey) {
		this.signKey = signKey;
	}
	public void setCertVerifyKey(PublicKey certVerifyKey) {
		this.certVerifyKey = certVerifyKey;
	}
	public void setPrivateKeyServer(PrivateKey privateKeyServer) {
		this.privateKeyServer = privateKeyServer;
	}
	public void setSessionKey(SymmKey sessionKey) {
		this.sessionKey = sessionKey;
	}
	public void setNonceServer(Nonce nonceServer) {
		this.nonceServer = nonceServer;
		this.nonceServer.setNonce(nonceServer.getNonce());
	}
	public void setNonceDeleteTicket(Nonce nonceDeleteTicket) {
		this.nonceDeleteTicket = nonceDeleteTicket;
		this.nonceDeleteTicket.setNonce(nonceDeleteTicket.getNonce());
	}
	public void setTicketID(ID ticketID) {
		this.ticketID = ticketID;
		this.ticketID.setId(ticketID.getId());
	}
	public void setIssuedTicketIDs(ListOfID issuedTicketIDs) {
		this.issuedTicketIDs = issuedTicketIDs;
	}
	public void setState(byte state) {
		this.state = state;
	}

	public void setTickets(ListOfCardTicket tickets) {
		this.tickets = tickets;
	}

	/*PROTECTED REGION ID(generated.eTicket.server.eTicketServer.ETicketServer) ENABLED START*/

	public ID generateID() {

		return once.eTicket.server.eTicketServer.Manual_ETicketServer
				.generateID(this);

	}
	public boolean pay(CCInfo ccInfo) {

		return once.eTicket.server.eTicketServer.Manual_ETicketServer.pay(
				ccInfo, this);

	}
	public Ticket createTicket(ID ticketId, TicketInfo ticketInfo) {

		return once.eTicket.server.eTicketServer.Manual_ETicketServer
				.createTicket(ticketId, ticketInfo, this);

	}
	/*PROTECTED REGION END*/

	public void startService() throws java.lang.Exception {

		endpoint = Endpoint.publish(
				"http://localhost:8000/ETicketServerService", this);

	}
	public void stopService() {
		endpoint.stop();

	}
	public static void main(String[] args) throws java.lang.Exception {
		ETicketServer s = new ETicketServer();
		s.startService();
		JOptionPane.showMessageDialog(null, "Server beenden");
		s.stopService();
	}

	protected void stop() throws ServiceException {

		throw new ServiceException("ETicketServer");
	}

}
