package generated.eTicket.terminal.inspectorDevice;

//import swt.simulation.SmartCardSimulation;
import swt.terminal.HandleResult;
import swt.util.ByteArray;
import swt.util.HexString;
import de.uniaugsburg.swt.javacard.terminalinterface.SWTReader;
import de.uniaugsburg.swt.javacard.terminalinterface.SWTReaderException;
import javax.xml.ws.BindingProvider;

public class InspectorDevice {

	private byte state;

	private Secret pin;

	private PrivateKey privateKeyTerminal;

	private PublicKey publicKeyTerminal;

	private PublicKey certVerifyKey;

	private PublicKey publicKeyCard;

	private SymmKey sessionKey;

	private Nonce terminalNonce;

	private ID ticketID;

	private ListOfID stampedTicketIDs;

	private Userinterface userinterface;
	public void setUserinterface(Userinterface u) {
		this.userinterface = u;
	}

	// saved prev message
	public byte[] prevMsg;
	public Usermessage prevUMsg;

	public final Gate dummyGate = new Gate() {
		@Override
		public void callConstraints() {
			/* do nothing */
		}

		@Override
		public byte[] transmit(InspectorDevice terminal, byte[] apdu, int port)
				throws TerminalException, SWTReaderException {
			return terminal.transmit(apdu, port);
		}

		@Override
		public void finishProtocol(InspectorDevice terminal) {
			/* do nothing */
		}

	};
	public Gate currentGate = dummyGate;

	/**
	 * Resets the current gate to a default dummy gate
	 */
	public void resetGate() {
		currentGate = dummyGate;
	}

	/**
	 * Sets the current gate
	 */
	public void setGate(Gate gate) {
		currentGate = gate;
	}

	public InspectorDevice() throws java.lang.Exception {

		state = StateTerminal.IDLE;
		pin = new Secret();
		privateKeyTerminal = new PrivateKey();
		publicKeyTerminal = new PublicKey();
		certVerifyKey = new PublicKey();
		publicKeyCard = new PublicKey();
		sessionKey = new SymmKey();
		terminalNonce = new Nonce();
		ticketID = new ID();
		stampedTicketIDs = new ListOfID();

	}

	public InspectorDevice(PublicKey certVerifyKey,
			PrivateKey privateKeyTerminal, PublicKey publicKeyTerminal,
			byte state) throws java.lang.Exception {

		this.certVerifyKey = certVerifyKey;
		this.privateKeyTerminal = privateKeyTerminal;
		this.publicKeyTerminal = publicKeyTerminal;
		this.state = state;

		pin = new Secret();
		publicKeyCard = new PublicKey();
		sessionKey = new SymmKey();
		terminalNonce = new Nonce();
		ticketID = new ID();
		stampedTicketIDs = new ListOfID();

	}

	private void sendMsg(Message msg) {

		currentGate.callConstraints();
		userinterface.send((Usermessage) msg);

	}

	private void sendMsg(Message msg, int port) throws java.lang.Exception {

		sendMsg(msg, port, false, false);

	}
	private void sendMsg(Message msg, int port, Boolean openSessionBeforeSend,
			Boolean closeSessionAfterReceive) throws java.lang.Exception {

		currentGate.callConstraints();

		Message response;

		switch (port) {
			case Ports.InspectorDevice2ETicketCard_default :
				response = sendCardMsg(msg, port);
				if (response != null)
					processMessage(response);
				break;

			default :
				stop();
		}
	}

	private Message sendCardMsg(Message msg, int port)
			throws java.lang.Exception {
		byte[] encMsg = new byte[LengthConstants.MAX_ENCODING_LENGTH_MESSAGES];
		short encMsgLen = Coding.getInstance().encode((Codeable) msg, encMsg);
		encMsg = ByteArray.subarray(encMsg, 0, encMsgLen);
		byte[] apdu = ByteArray.append(new byte[]{0, 2, 0, 0, 0}, encMsg);
		byte[] response = currentGate.transmit(this, apdu, port);

		if (response.length == 2) {
			if (!(response[0] == (byte) 0x90 && response[1] == (byte) 0x00))
				throw new TerminalException("Cardexception: "
						+ HandleResult.interpretSW(response));
			return null;
		} else {
			prevMsg = response;

			Coding.getInstance().decodeInit();
			Codeable cResponse = Coding.getInstance().decode(response);
			return (Message) cResponse;
		}
	}

	/**
	 * Processes a raw data stream by trying to decode data to Message.
	 * Needed for testcases.
	 * @param data
	 */
	public void processRawMessage(byte[] data) {
		prevMsg = data;
		try {
			Coding.getInstance().decodeInit();
			Codeable cResponse = Coding.getInstance().decode(data);
			processMessage((Message) cResponse);
		} catch (java.lang.Exception e) {
			System.err.println("Failed decoding data in processRawMessage");
		}

	}

	private void processMessage(Message inmsg) throws java.lang.Exception {
		currentGate.callConstraints();

		switch (inmsg.getCode()) {
			case Code.UCARDAUTH :
				processUCardAuth((UCardAuth) inmsg);
				break;
			case Code.CARDAUTHRES :
				processCardAuthRes((CardAuthRes) inmsg);
				break;
			case Code.USTAMPTICKET :
				processUStampTicket((UStampTicket) inmsg);
				break;
			case Code.STAMPTICKETSUCCESSFUL :
				processStampTicketSuccessful((StampTicketSuccessful) inmsg);
				break;
			case Code.USHOWALLITICKETS :
				processUShowAllITickets((UShowAllITickets) inmsg);
				break;
			case Code.RETURNALLITICKETS :
				processReturnAllITickets((ReturnAllITickets) inmsg);
				break;
			case Code.USETPASSWORD :
				processUSetPassword((USetPassword) inmsg);
				break;
			default :
				stop();
		}

		currentGate.callConstraints();
		currentGate.finishProtocol(this);
	}

	public void process(Message msg) throws TerminalException {
		currentGate.callConstraints();

		try {
			if (msg instanceof Usermessage)
				prevUMsg = (Usermessage) msg;
			processMessage(msg);
		} catch (java.lang.Exception e) {
			// e.printStackTrace();
			if (e instanceof TerminalException)
				throw (TerminalException) e;
			else
				throw new TerminalException(
						"Error when calling process method: " + e);
		}

		currentGate.callConstraints();
		currentGate.finishProtocol(this);
	}

	private final static int MAX_SPLIT_LENGTH = 50;
	private final static int MAX_APDU_LENGTH = 254;
	private final static int APDU_HEADER_LENGTH = 5;
	private final static int APDU_DATA_LENTH = MAX_APDU_LENGTH
			- APDU_HEADER_LENGTH;
	private SWTReader[] reader = new SWTReader[2];

	public void initReader(SWTReader reader, int port) {
		this.reader[port] = reader;
	}

	public byte[] encode(Codeable c) throws java.lang.Exception {
		byte[] encMsg = new byte[LengthConstants.MAX_ENCODING_LENGTH_MESSAGES];
		short encMsgLen = Coding.getInstance().encode(c, encMsg);
		return ByteArray.subarray(encMsg, 0, encMsgLen);
	}

	public byte[] transmit(byte[] apdu, int port) throws TerminalException,
			SWTReaderException {
		byte[] received = null;
		if (apdu.length <= MAX_APDU_LENGTH) {
			apdu[4] = (byte) (apdu.length - 5);
			received = reader[port].transmit(apdu);
			if (received.length == 2)
				return received;
			received = ByteArray.subarray(received, 0, received.length - 2);
		} else {
			byte[] splittedApdu;
			byte[] apduHeader = ByteArray.subarray(apdu, 0, APDU_HEADER_LENGTH);
			apdu = ByteArray.subarray(apdu, APDU_HEADER_LENGTH);
			int splitCount = apdu.length / APDU_DATA_LENTH;
			if (apdu.length % APDU_DATA_LENTH != 0)
				splitCount++;
			for (int i = 0; i < splitCount; i++) {
				if (i == splitCount - 1) {
					splittedApdu = ByteArray
							.subarray(apdu, i * APDU_DATA_LENTH);
					apduHeader[2] = (byte) 0;
					apduHeader[4] = (byte) splittedApdu.length;
					splittedApdu = ByteArray.append(apduHeader, splittedApdu);
					received = reader[port].transmit(splittedApdu);
					if (received.length == 2)
						return received;
					received = ByteArray.subarray(received, 0,
							received.length - 2);
				} else {
					splittedApdu = ByteArray.subarray(apdu,
							i * APDU_DATA_LENTH, APDU_DATA_LENTH);
					apduHeader[2] = (byte) 1;
					apduHeader[4] = (byte) splittedApdu.length;
					splittedApdu = ByteArray.append(apduHeader, splittedApdu);
					received = reader[port].transmit(splittedApdu);
					if (!(received[0] == (byte) 0x90 && received[1] == (byte) 0x00)) {
						throw new TerminalException(
								"Bad answer from card: part of a long APDU was not answered with OK (0x9000).");
					}
				}
			}
		}

		for (int i = 0; i < MAX_SPLIT_LENGTH; i++) {
			if (received[0] == 0) {
				received = ByteArray.subarray(received, 1);
				return received;
			} else if (received[0] == 1) {
				received = ByteArray.subarray(received, 1);
				byte[] getMore = new byte[]{0, 6, 0, 0, 0, 1, 1};
				byte[] temp_received = reader[port].transmit(getMore);

				temp_received = ByteArray.subarray(temp_received, 0,
						temp_received.length - 2);
				received = ByteArray.append(temp_received[0], received);
				temp_received = ByteArray.subarray(temp_received, 1);
				received = ByteArray.append(received, temp_received);
			} else
				throw new TerminalException(
						"Bad answer from card: answer to a long APDU must begin with 0 or 1.");
		}
		return received;
	}

	/**
	 * This method initializes the card with the given parameters that correspond 
	 * to the card attributes. A null value will be ignored by the card.
	 * @throws Exception
	 */
	public void initCard_ETicketCard(int port, CCInfo ccInfo,
			SignedData certCard, ID id, Secret pin, PrivateKey privateKeyCard,
			PublicKey publicKeyCard, PublicKey publicKeyServer, byte state)
			throws java.lang.Exception {
		byte cardCode = Code.ETICKETCARD;
		byte[] apdu = {0, 4, 0, 0, 0, cardCode};
		apdu = ByteArray.append(apdu, (ccInfo != null)
				? encode(ccInfo)
				: new byte[]{Code.IGNORE});

		apdu = ByteArray.append(apdu, (certCard != null)
				? encode(certCard)
				: new byte[]{Code.IGNORE});

		apdu = ByteArray.append(apdu, (id != null)
				? encode(id)
				: new byte[]{Code.IGNORE});

		apdu = ByteArray.append(apdu, (pin != null)
				? encode(pin)
				: new byte[]{Code.IGNORE});

		apdu = ByteArray.append(apdu, (privateKeyCard != null)
				? encode(privateKeyCard)
				: new byte[]{Code.IGNORE});

		apdu = ByteArray.append(apdu, (publicKeyCard != null)
				? encode(publicKeyCard)
				: new byte[]{Code.IGNORE});

		apdu = ByteArray.append(apdu, (publicKeyServer != null)
				? encode(publicKeyServer)
				: new byte[]{Code.IGNORE});

		apdu = ByteArray.append(apdu, new byte[]{Code.BYTE, state});
		transmit(apdu, port);
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	private void processUCardAuth(UCardAuth inmsg) throws java.lang.Exception {

		if (state == StateTerminal.getIDLE()) {
			setTerminalNonce(Nonce.generateNonce());
			setState(StateTerminal.getEXP_CARD_AUTH());
			sendMsg(new CardAuth(publicKeyTerminal),
					Ports.InspectorDevice2ETicketCard_default);
		} else {
			setState(StateTerminal.getIDLE());
			stop();
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	private void processCardAuthRes(CardAuthRes inmsg)
			throws java.lang.Exception {
		PublicKey publicKeyCard = inmsg.getPublicKeyCard();
		SignedData certCard = inmsg.getCertCard();
		EncDataAsymm enc = inmsg.getMsg();
		if (state == StateTerminal.getEXP_CARD_AUTH()) {
			Certificate authData = new Certificate(publicKeyCard);
			boolean correctCert = SignedData.verify(certVerifyKey, certCard,
					authData);
			if (correctCert) {
				this.setPublicKeyCard(publicKeyCard);
				CardAuthResMsg msg = (CardAuthResMsg) (EncDataAsymm.decrypt(
						privateKeyTerminal, enc));
				setSessionKey(msg.getSessionKey());
				setState(StateTerminal.getAUTHENTICATED());
				sendMsg(new UCardAuthSuccessful());
			} else {
				setState(StateTerminal.getIDLE());
				stop();
			}
		} else {
			setState(StateTerminal.getIDLE());
			stop();
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	private void processUStampTicket(UStampTicket inmsg)
			throws java.lang.Exception {
		ID ticketId = inmsg.getTicketId();
		if (state == StateTerminal.getAUTHENTICATED()) {
			Stamp stamp = stamp();
			StampTicketMsg std = new StampTicketMsg(ticketId, pin, stamp);
			EncDataAsymm enc = EncDataAsymm.encrypt(publicKeyCard, std);
			setTicketID(ticketId);
			setState(StateTerminal.getIDLE());
			sendMsg(new StampTicket(enc),
					Ports.InspectorDevice2ETicketCard_default);
		} else {
			setState(StateTerminal.getIDLE());
			stop();
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	private void processStampTicketSuccessful(StampTicketSuccessful inmsg)
			throws java.lang.Exception {

		stampedTicketIDs.add(ticketID);
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	private void processUShowAllITickets(UShowAllITickets inmsg)
			throws java.lang.Exception {

		if (state == StateTerminal.getAUTHENTICATED()) {
			sendMsg(new ShowAllITickets(),
					Ports.InspectorDevice2ETicketCard_default);
		} else {
			setState(StateTerminal.getIDLE());
			stop();
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	private void processReturnAllITickets(ReturnAllITickets inmsg)
			throws java.lang.Exception {
		EncDataSymm enc = inmsg.getMsg();
		ReturnAllITicketsMsg msg = (ReturnAllITicketsMsg) (EncDataSymm.decrypt(
				sessionKey, enc));
		sendMsg(new UReturnAllITickets(msg.getTickets()));
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	private void processUSetPassword(USetPassword inmsg)
			throws java.lang.Exception {
		Secret pin = inmsg.getPin();
		this.setPin(pin);
	}

	public byte getState() {
		return state;
	}

	public Secret getPin() {
		return pin;
	}

	public PrivateKey getPrivateKeyTerminal() {
		return privateKeyTerminal;
	}

	public PublicKey getPublicKeyTerminal() {
		return publicKeyTerminal;
	}

	public PublicKey getCertVerifyKey() {
		return certVerifyKey;
	}

	public PublicKey getPublicKeyCard() {
		return publicKeyCard;
	}

	public SymmKey getSessionKey() {
		return sessionKey;
	}

	public Nonce getTerminalNonce() {
		return terminalNonce;
	}

	public ID getTicketID() {
		return ticketID;
	}

	public ListOfID getStampedTicketIDs() {
		return stampedTicketIDs;
	}

	public void setState(byte state) {
		this.state = state;
	}

	public void setPin(Secret pin) {
		this.pin = pin;
	}
	public void setPrivateKeyTerminal(PrivateKey privateKeyTerminal) {
		this.privateKeyTerminal = privateKeyTerminal;
	}
	public void setPublicKeyTerminal(PublicKey publicKeyTerminal) {
		this.publicKeyTerminal = publicKeyTerminal;
	}
	public void setCertVerifyKey(PublicKey certVerifyKey) {
		this.certVerifyKey = certVerifyKey;
	}
	public void setPublicKeyCard(PublicKey publicKeyCard) {
		this.publicKeyCard = publicKeyCard;
	}
	public void setSessionKey(SymmKey sessionKey) {
		this.sessionKey = sessionKey;
	}
	public void setTerminalNonce(Nonce terminalNonce) {
		this.terminalNonce = terminalNonce;
	}
	public void setTicketID(ID ticketID) {
		this.ticketID = ticketID;
	}
	public void setStampedTicketIDs(ListOfID stampedTicketIDs) {
		this.stampedTicketIDs = stampedTicketIDs;
	}

	/*PROTECTED REGION ID(generated.eTicket.terminal.inspectorDevice.InspectorDevice) ENABLED START*/

	public Stamp stamp() {

		return once.eTicket.terminal.inspectorDevice.Manual_InspectorDevice
				.stamp(this);

	}
	/*PROTECTED REGION END*/

	protected void stop() throws TerminalException {
		currentGate.callConstraints();
		throw new TerminalException("InspectorDevice");
	}

}
